@@ -1,4 +1,4 @@
-/* Copyright (C) 2003 Nikos Chantziaras.
+/* Copyright (C) 2004 Nikos Chantziaras.
  *
  * This file is part of the QTads program.  This program is free
  * software; you can redistribute it and/or modify it under the
@@ -20,15 +20,38 @@
 #ifndef CONFIG_H
 #define CONFIG_H
 
-#define QTADSVERSION "QTads 1.3a"
+#include <qconfig.h>
+
+#define QTADSVERSION "QTads 1.7"
+
+/* Recent versions of Qt define Q_BYTE_ORDER.  If defined, we use it to
+ * decide which version of the endian-related Tads routines to use.  If
+ * Q_BYTE_ORDER is not defined, we always use the PowerPC version, which
+ * (desptite the name) are suited for both big-endian as well as little-
+ * endian machines.  _M_IX86 is the little-endian version, and is faster
+ * on Intel-compatibles than the generic PPC version.
+ */
+ /*
+#ifdef Q_BYTE_ORDER
+#if Q_BYTE_ORDER == Q_LITTLE_ENDIAN
+#define _M_PPC
+#else
+#define _M_PPC
+#endif
+#else
+#define _M_PPC
+#endif
+*/
 
 /* G++ (GNU C++) versions older than 3.0 didn't provide the 'and',
  * 'not' and 'or' keywords, so we define them as macros.
  */
+#ifdef __GNUC__
 #if __GNUC__ == 2
 #define not !
 #define and &&
 #define or ||
 #endif
+#endif
 
 #endif /* CONFIG_H */
