@@ -1,4 +1,4 @@
-/* Copyright (C) 2003 Nikos Chantziaras.
+/* Copyright (C) 2004 Nikos Chantziaras.
  *
  * This file is part of the QTads program.  This program is free
  * software; you can redistribute it and/or modify it under the
@@ -30,14 +30,21 @@
 int
 memicmp( const char* s1, const char* s2, size_t len )
 {
-	char x1[len];
-	char x2[len];
+	// TODO: Hmm, are variable-size arrays really not in ISO C++?
+	//char x1[len];
+	//char x2[len];
+
+	char* x1 = new char[len];
+	char* x2 = new char[len];
 
 	for (size_t i = 0; i < len; ++i) {
 		x1[i] = std::tolower(s1[i]);
 		x2[i] = std::tolower(s2[i]);
 	}
-	return std::memcmp(x1, x2, len);
+	int ret = std::memcmp(x1, x2, len);
+	delete[] x1;
+	delete[] x2;
+	return ret;
 }
 
 
