@@ -1,4 +1,4 @@
-/* Copyright (C) 2003 Nikos Chantziaras.
+/* Copyright (C) 2004 Nikos Chantziaras.
  *
  * This file is part of the QTads program.  This program is free
  * software; you can redistribute it and/or modify it under the
@@ -34,21 +34,22 @@
  * are static functions (they aren't exported).
  *
  * Again: No functions are allowed in here that aren't listed in
- * "osifc.h"!
+ * "osifc.h" except helper functions.
  */
 
-/* A note about Multimedia Tads Character Entities
+/* A note about Multimedia-Tads Character Entities
  * ===============================================
  *
  * QTads is able to display Multimedia Tads Character Entities (like
  * "&rdquo;" or "&emdash;") even in Tads 2 games.  These characters are
  * nothing more than Unicode values.  But Tads 2 does not support
- * Unicode (Tads 3 does).  To work around this, QTads uses UTF-8
- * encoded Unicode for output.  The problem with UTF-8 is that it only
- * contains 7-bit ASCII as a subset, so 8-bit character sets like
- * Latin1 don't work.  Because of this, QTads uses a modification of
- * UTF-8 that is able to include a full 8-bit character set as a subset.
- * A char* string with this encoding looks like this:
+ * Unicode (Tads 3 does).  To work around this, QTads uses a "hacked"
+ * version of UTF-8 encoded Unicode for output.  The problem with UTF-8
+ * is that it only contains 7-bit ASCII as a subset, so 8-bit character
+ * sets like Latin1 don't work.  That's why QTads uses a modification
+ * of UTF-8 that is able to include a full 8-bit character set as a
+ * subset.  This "hacked" UTF-8 encoding is called QTF.  A QTF-encoded
+ * 'char[]' looks like this:
  *
  *   If a byte in this string is 0, then this is where the string ends
  *   (as usual; in this aspect, it's like regular UTF-8).
@@ -63,9 +64,8 @@
  *
  *     [1] [byte 1] [byte 2 (if any)] [byte 3 (if any)]
  *
- * In the source code, we refer to this special encoding as "QTF".  The
- * function qtf2QString() (defined below) can be used to convert these
- * specially encoded strings to regular Unicode QStrings.
+ * The function qtf2QString() (defined below) can be used to convert
+ * these specially encoded strings to regular Unicode QStrings.
  *
  * Note that this method has a potential drawback; if a game tries to
  * display a character with value 1, this character will be
@@ -78,7 +78,13 @@
  * interpreter in this world knows how to display this character.
  *
  * Note that QTF is not used in Tads 3, since Tads 3 already supports
- * Unicode and there's no need for this kind of hack.
+ * Unicode and there's no need for this kind of hack; in Tads 3, we
+ * simply use UTF-8.
+ *
+ * Note that we can't allow transcript files to contain QTF characters.
+ * We keep a list of every opened logfile so that we can translate the
+ * QTF strings to the local character set when output is performed on a
+ * logfile.
  */
 
 #include "config.h"
@@ -92,14 +98,21 @@
 #include <qtextcodec.h>
 
 #include <vector>
+#include <list>
+#include <algorithm>
 #include <cstring>
 #include <cstdio>
 #include <cctype>
 
 #include "os.h"
 
+#include "qtadstypes.h"
+#include "qtadswarning.h"
+#include "qtadscharmap.h"
+
 #ifndef HTMLQT
 #include "qtadsio.h"
+#include "qtadssettings.h"
 #include "install_dirs.h"
 #endif
 
@@ -109,22 +122,45 @@
  */
 static int status_mode = 0;
 
+/* We use these to tell os_getc_raw() if it should use a timeout.  If
+ * timedInputEvent is true, timedInputEventAmount is the timeout in
+ * milliseconds.  After os_getc_raw() has returned, inputHasTimedOut
+ * should be checked to see if the event timed out.
+ */
+static bool timedInputEvent = false;
+static unsigned long timedInputEventAmount = 0;
+static bool inputHasTimedOut = false;
+
+/* We keep a list of every logfile (transcript) that is currently open
+ * by the Tads2 VM.  We use this to translate the QTF-encoded text into
+ * the system's local character set.
+ */
+static std::list<osfildef*> logFiles;
+
+/* Since the T3VM isn't running when we're in Tads2 mode, we can't make
+ * use of the T3VM character mapping API for Tads2 games; we need our
+ * own mapper.
+ */
+static QTadsCharmapToLocal logfileMapper;
+
 
 /* Converts a QTF encoded string to a QString.
+ * If 'strlen' != 0 then 's' is not null terminated.
  */
 static inline
 QString
-qtf2QString( const char* s )
+qtf2QString( const char* s, size_t strlen = 0 )
 {
 	QString res;
-	const size_t len = std::strlen(s);
+	const size_t len = strlen == 0 ? std::strlen(s) : strlen;
+
 	for (size_t i = 0; i < len; ++i) {
 		if (s[i] == 1) {
 			// A multibyte UTF-8 character follows.
 			// Determine its length and copy it to the
 			// result.
 			char tmp[4];
-			unsigned int c = (unsigned int)(unsigned char)s[i+1];
+			unsigned c = static_cast<unsigned char>(s[i+1]);
 
 			// The result of the switch is the length in
 			// bytes.  The code is ugly, but very fast.
@@ -133,8 +169,7 @@ qtf2QString( const char* s )
 			//   111xxxxx: Character is 3 bytes long.
 			//   110xxxxx: Character is 2 bytes long.
 			//   0xxxxxxx: Character is 1 byte long.
-			switch (1 + ((((c & 0x80) >> 7) | ((c & 0x80) >> 6))
-				     & (1 + ((c & 0x20) >> 5))))
+			switch (1 + ((((c & 0x80) >> 7) | ((c & 0x80) >> 6)) & (1 + ((c & 0x20) >> 5))))
 			{
 			  case 1:
 				tmp[0] = s[++i];
@@ -175,21 +210,55 @@ qtf2QString( const char* s )
  * text and binary files, since (as far as I know) the standard
  * functions always do the right thing; a "b" in the mode string is
  * ignored on systems that treat text and binary files the same (like
- * most/all Unix-systems).
+ * most/all POSIX-systems).
  */
 
+/* Open text file for writing.
+ */
+osfildef*
+osfopwt( const char* fname, os_filetype_t filetype )
+{
+	Q_ASSERT(fname != 0);
+
+	osfildef* fp = std::fopen(fname, "w");
+	if (fp != 0 and not QTadsIO::t3Mode() and filetype == OSFTLOG) {
+		// We're in Tads2 mode, and it's a logfile;
+		// store the pointer in the log-list.
+		::logFiles.push_back(fp);
+	}
+	return fp;
+}
+
+
 /* Open text file for reading and writing.
  */
 osfildef*
-osfoprwt( const char* fname, os_filetype_t )
+osfoprwt( const char* fname, os_filetype_t filetype)
 {
 	Q_ASSERT(fname != 0);
 
 	osfildef* fp = std::fopen(fname, "a+");
-	// If the operation was successful, position the seek pointer
-	// to the beginning of the file.
 	if (fp != 0) {
+		// Position the seek pointer to the beginning of the file.
 		std::rewind(fp);
+		if (not QTadsIO::t3Mode() and filetype == OSFTLOG) {
+			::logFiles.push_back(fp);
+		}
+	}
+	return fp;
+}
+
+
+/* Open text file for reading/writing.  If the file already exists,
+ * truncate the existing contents.  Create a new file if it doesn't
+ * already exist.
+ */
+osfildef*
+osfoprwtt( const char* fname, os_filetype_t filetype )
+{
+	osfildef* fp = std::fopen((fname),"w+");
+	if (fp != 0 and not QTadsIO::t3Mode() and filetype == OSFTLOG) {
+		::logFiles.push_back(fp);
 	}
 	return fp;
 }
@@ -198,9 +267,14 @@ osfoprwt( const char* fname, os_filetype_t )
 /* Open binary file for reading/writing.
  */
 osfildef*
-osfoprwb( const char* fname, os_filetype_t )
+osfoprwb( const char* fname, os_filetype_t
+#ifdef DEBUG
+	filetype
+#endif
+)
 {
 	Q_ASSERT(fname != 0);
+	Q_ASSERT(filetype != OSFTLOG);
 
 	osfildef* fp = std::fopen(fname, "a+b");
 	if (fp != 0) {
@@ -210,6 +284,20 @@ osfoprwb( const char* fname, os_filetype_t )
 }
 
 
+/* Close a file.
+ */
+void
+osfcls( osfildef* fp )
+{
+	// Remove the pointer from the logfiles-list.  If it's not in
+	// the list, nothing happens; no need to check it.
+	::logFiles.remove(fp);
+
+	// Close the file.
+	std::fclose(fp);
+}
+
+
 /* Access a file - determine if the file exists.
  */
 int
@@ -221,6 +309,26 @@ osfacc( const char* fname )
 }
 
 
+/* Print a counted-length string (which might not be null-terminated)
+ * to a file.
+ */
+void
+os_fprint( osfildef* fp, const char* str, size_t len )
+{
+	Q_ASSERT(fp != 0);
+	Q_ASSERT(str != 0);
+
+	if (QTadsIO::t3Mode()
+	    or std::find(::logFiles.begin(), ::logFiles.end(), fp) == ::logFiles.end())
+	{
+		std::fprintf(fp, "%.*s", len, str);
+	} else {
+		const QCString& tmp = qtf2QString(str, len).utf8();
+		::logfileMapper.write_file(fp, tmp, tmp.length());
+	}
+}
+
+
 /* Write a null-terminated string to a text file.
  */
 void
@@ -229,7 +337,14 @@ os_fprintz( osfildef* fp, const char* str )
 	Q_ASSERT(fp != 0);
 	Q_ASSERT(str != 0);
 
-	std::fprintf(fp, "%s", str);
+	if (QTadsIO::t3Mode()
+	    or std::find(::logFiles.begin(), ::logFiles.end(), fp) == ::logFiles.end())
+	{
+		std::fprintf(fp, "%s", str);
+	} else {
+		const QCString& tmp = qtf2QString(str).utf8();
+		::logfileMapper.write_file(fp, tmp, tmp.length());
+	}
 }
 
 
@@ -264,50 +379,16 @@ os_exeseek( const char*, const char* )
 }
 
 
-/* --------------------------------------------------------------------
- * External, dynamically-loaded code.
- *
- * This Tads-feature is deprecated.  I don't see why it should be
- * supported.  Did any game actually use it?
- */
-
-/* Load an external function from a file given the name of the file.
- */
-int
-(*os_exfil(const char*))(void*)
-{
-	qWarning("Dynamically-loaded code is not supported in QTads");
-	return static_cast<int (*)(void*)>(0);
-}
-
-
-/* Load an external function from an open file.
- */
-int
-(*os_exfld(osfildef*, unsigned))(void*)
-{
-	qWarning("Dynamically-loaded code is not supported in QTads");
-	return static_cast<int (*)(void*)>(0);
-}
-
-
-/* Call an external function.
- */
-int
-os_excall(int (*)(void*), void*)
-{
-	qWarning("Dynamically-loaded code is not supported in QTads");
-	return 0;
-}
-
-
 /* --------------------------------------------------------------------
  */
 
 /* Look for a file in the standard locations: current directory,
- * program directory, TADS path.
+ * program directory, PATH.
  *
- * TODO: For now, we only look in the current directory.
+ * We only look in the current directory, which is the game's
+ * directory.
+ *
+ * TODO: I don't know how useful it'll be to also search the PATH.
  */
 int
 os_locate( const char* fname, int /*flen*/, const char* /*arg0*/, char* buf, size_t bufsiz )
@@ -315,6 +396,7 @@ os_locate( const char* fname, int /*flen*/, const char* /*arg0*/, char* buf, siz
 	Q_ASSERT(fname != 0);
 	Q_ASSERT(buf != 0);
 
+	// Current directory.
 	if (QFile::exists(fname)) {
 		Q_ASSERT(bufsiz > std::strlen(QFileInfo(fname).absFilePath().ascii()));
 
@@ -322,6 +404,8 @@ os_locate( const char* fname, int /*flen*/, const char* /*arg0*/, char* buf, siz
 		buf[bufsiz - 1] = '\0';
 		return 1;
 	}
+
+	// Not found.
 	return 0;
 }
 
@@ -355,6 +439,9 @@ osfdel_temp( const char* fname )
 	Q_ASSERT(fname != 0);
 
 	if (fname[0] == '\0' or QFile::remove(fname)) {
+		// If fname was empty, it has been already deleted
+		// automatically by fclose().  If fname was not empty,
+		// QFile::remove has taken care of deleting the file.
 		return 0;
 	}
 	// QFile::remove() failed.
@@ -384,7 +471,7 @@ os_defext( char* fn, const char* ext )
 /* Unconditionally add an extention to a filename.
  *
  * TODO: Find out if there are systems that don't use the dot as the
- * extension separator.  (Only systems Qt supports of course.)
+ * extension separator.  (Only systems supported by Qt of course.)
  */
 void
 os_addext( char* fn, const char* ext )
@@ -469,24 +556,17 @@ os_rand( long* val )
 {
 	Q_ASSERT(val != 0);
 
-	static bool initialized = false;
-	std::time_t t = std::time(0);
+	time_t t = time(0);
 
-	if (not initialized) {
-		if (t == static_cast<std::time_t>(-1)) {
-			std::srand(std::rand());
-		} else {
-			std::srand(static_cast<unsigned int>(t));
-		}
-		initialized = true;
+	if (t == static_cast<time_t>(-1)) {
+		std::srand(std::rand());
+	} else {
+		std::srand(static_cast<unsigned int>(t));
 	}
 
 	// Generate a random number by using high-order bits, because
 	// on some systems the low-order bits aren't very random.
-	// Note that this is Voodoo Programming for me; I simply used
-	// the `rand' manpage as a guide.
-	*val = 1 + static_cast<long>(static_cast<long double>(65535)
-				      * std::rand() / (RAND_MAX + 1.0));
+	*val = 1 + static_cast<long>(static_cast<long double>(65535) * std::rand() / (RAND_MAX + 1.0));
 }
 
 
@@ -505,18 +585,23 @@ os_printz( const char* str )
 	Q_ASSERT(str != 0);
 
 	if (str[0] == '\0' or ::status_mode > 1) {
+		// Nothing to print or status line mode forbids
+		// printing.
 		return;
 	}
 
-	const QString tmp(QTadsIO::t3Mode() ? QString::fromUtf8(str) : ::qtf2QString(str));
+	const QString& tmp = QTadsIO::t3Mode() ? QString::fromUtf8(str) : ::qtf2QString(str);
 
 	if (::status_mode == 0) {
+		// Status line more is 0; print to main window.
 		QTadsIO::print(tmp);
 	} else {
+		// Prepare to print to status line.
 		QString res;
 		bool ret = false;
 		uint i = 0;
 
+		// Skip any newline-characters.
 		if (tmp[0] == '\n') {
 			while (i < tmp.length() and tmp[i] == '\n') {
 				++i;
@@ -525,11 +610,17 @@ os_printz( const char* str )
 
 		while (i < tmp.length() and ::status_mode == 1) {
 			if (tmp[i] == '\r') {
+				// The "\r" sequence deletes the preceding
+				// text.
 				res = "";
 				ret = true;
 			} else if (tmp[i] == '\n') {
+				// A newline sets the status line mode
+				// to 2, which means that subsequent
+				// text should be ignored.
 				::status_mode = 2;
 			} else {
+				// It's a normal character; keep it.
 				res += tmp[i];
 			}
 			++i;
@@ -572,10 +663,7 @@ os_get_status()
 void
 os_score( int score, int turncount )
 {
-	QString res(QString::number(score));
-	res += "/";
-	res += QString::number(turncount);
-	QTadsIO::scorePrint(res);
+	QTadsIO::scorePrint(QString::number(score) + "/" + QString::number(turncount));
 }
 #endif
 
@@ -635,7 +723,7 @@ os_set_text_attr( int attr )
 /* Use plain ascii mode for the display.
  *
  * We don't provide this.  It's possible to emulate it, but there's no
- * point in doing so.
+ * point in doing it.
  */
 #ifndef HTMLQT
 void
@@ -812,6 +900,11 @@ os_gets( unsigned char* buf, size_t bufl )
 
 /* Read a character from the keyboard and return the low-level,
  * untranslated key code whenever possible.
+ *
+ * Note that we deviate from the required behavior, in that we can
+ * return 0 even when not an extended character is to be returned; but
+ * not when the VM is calling us.  Only when the operation has timed
+ * out.  (os_get_event() is implemented through this function.)
  */
 #ifndef HTMLQT
 int
@@ -821,12 +914,21 @@ os_getc_raw( void )
 	// returned 0, so this time we should return the extended
 	// key-code.
 	static bool done = true;
-	// QKeyEvent lacks a default ctor, so we provide some random
-	// values.
-	static QKeyEvent e(QEvent::KeyPress, 0, 0, 0);
+
+	// Construct a default key event.  We make it static since we
+	// might get called twice in order to return an "extended"
+	// character-code.
+	static QTadsKeyEvent e(false);
 
 	if (done) {
-		e = QTadsIO::getRawChar();
+		e = QTadsIO::getRawChar(::timedInputEvent, ::timedInputEventAmount);
+		if (e.hasTimedOut()) {
+			// The operation timed out.  Report it and return.
+			Q_ASSERT(::timedInputEvent == true);
+			::inputHasTimedOut = true;
+			return 0;
+		}
+		::inputHasTimedOut = false;
 		// Store an upper-case version in case it's an
 		// ALT+[KEY] combination.
 		int asc = std::toupper(e.ascii());
@@ -926,20 +1028,27 @@ os_waitc( void )
 
 
 /* Get an input event.
- *
- * TODO: Implement timeout.
  */
 #ifndef HTMLQT
 int
-os_get_event( unsigned long /*timeout*/, int use_timeout, os_event_info_t* info )
+os_get_event( unsigned long timeout, int use_timeout, os_event_info_t* info )
 {
 	if (use_timeout) {
-		// We don't support timed events (yet).
-		return OS_EVT_NOTIMEOUT;
+		::timedInputEvent = true;
+		::timedInputEventAmount = timeout;
+	} else {
+		::timedInputEvent = false;
 	}
 
 	// Wait for an input from the keyboard.
 	int res = os_getc_raw();
+	::timedInputEvent = false;
+
+	if (::inputHasTimedOut) {
+		// The operation timed out.
+		Q_ASSERT(res == 0);
+		return OS_EVT_TIMEOUT;
+	}
 
 	if (not QTadsIO::gameRunning) {
 		// For some reason, the game stopped executing.  Return
@@ -971,8 +1080,14 @@ os_get_event( unsigned long /*timeout*/, int use_timeout, os_event_info_t* info
  * TODO: Implement default-button.
  */
 int
-os_input_dialog( int icon_id, const char* prompt, int standard_button_set,
-		 const char** buttons, int button_count, int default_index, int cancel_index )
+os_input_dialog( int
+#ifdef DEBUG
+		// Suppress "unused parameter" warning when building
+		// the release version.
+		icon_id
+#endif
+		, const char* prompt, int standard_button_set, const char** buttons,
+		int button_count, int default_index, int cancel_index )
 {
 	Q_ASSERT(prompt != 0);
 	Q_ASSERT(icon_id == OS_INDLG_ICON_NONE or icon_id == OS_INDLG_ICON_WARNING
@@ -1055,7 +1170,7 @@ os_get_sys_clock_ms( void )
 {
 	static QTime zeroPoint(QTime::currentTime());
 	static int lastRet = -1;
-	static int wraps = 0;
+	static unsigned int wraps = 0;
 
 	int ret = zeroPoint.elapsed();
 
@@ -1064,7 +1179,9 @@ os_get_sys_clock_ms( void )
 		// 24 hours have passed since this function has been
 		// called for the first time.  It's unlikely that
 		// someone will run QTads for 24 hours, but better to
-		// be sure.
+		// be sure.  Note that the code *will* break if QTads
+		// is run for more than 11.767.033 years 251 days 20
+		// hours and 24 minutes.
 		++wraps;
 		zeroPoint.start();
 	}
@@ -1131,12 +1248,13 @@ os_init( int* /*argc*/, char* /*argv*/[], const char* /*prompt*/, char* /*buf*/,
 void
 os_uninit( void )
 {
+	// I can't think of anything to do here.
 }
 
 
 /* Pause prior to exit, if desired.
  *
- * TODO: Implement it.
+ * I don't think we need this.
  */
 #ifndef HTMLQT
 void
@@ -1152,6 +1270,7 @@ void
 os_term( int /*status*/ )
 {
 #ifndef HTMLQT
+	// Tell the main-loop to quit the app.
 	QTadsIO::quitApp = true;
 #endif
 }
@@ -1199,24 +1318,32 @@ os_set_save_ext( const char* )
 /* Translate a character from the HTML 4 Unicode character set to the
  * current character set used for display.
  *
- * We encode the character as QTF.
+ * We encode the character as QTF, unless the encoding maps directly to
+ * 7-bit ASCII, in which case no translation is needed.
+ *
+ * This routine is (or rather, *should*) get called only by the T2VM.
  */
 #ifndef HTMLQT
 void
 os_xlat_html4( unsigned int html4_char, char* result, size_t result_buf_len )
 {
 	Q_ASSERT(result != 0);
-	Q_ASSERT(result_buf_len > 4);
 	Q_ASSERT(not QTadsIO::t3Mode());
 
-	const QString& tmp = QString(QChar(html4_char));
-
-	result[0] = 1;
+	const QChar c(html4_char);
 	result[1] = '\0';
 
-	//std::strncat(result, tmp.utf8(), result_buf_len - std::strlen(tmp.utf8()) - 1);
-	std::strcat(result, tmp.utf8());
-	result[result_buf_len - 1] = '\0';
+	if (c.unicode() < 128) {
+		// The character's unicode value maps directly to
+		// ASCII; store it as-is.
+		result[0] = c;
+	} else {
+		// The character does not map to ASCII; encode it as QTF.
+		result[0] = 1;
+		Q_ASSERT(result_buf_len > QString(c).utf8().length() + 1);
+		std::strcat(result, QString(c).utf8());
+		result[result_buf_len - 1] = '\0'; // Just in case.
+	}
 }
 #endif
 
