@@ -1,4 +1,4 @@
-/* Copyright (C) 2003 Nikos Chantziaras.
+/* Copyright (C) 2004 Nikos Chantziaras.
  *
  * This file is part of the QTads program.  This program is free
  * software; you can redistribute it and/or modify it under the
@@ -24,7 +24,6 @@
 #include <qstring.h>
 #include <qapplication.h>
 #include <qeventloop.h>
-#include <qpopupmenu.h>
 #include <qcursor.h>
 
 #include "qtadstypes.h"
@@ -38,7 +37,8 @@ const QString MORE_TEXT(QObject::tr("*** More *** (Press SPACE to continue)"));
 QTadsGameWindow::QTadsGameWindow( QWidget* parent, const char* name )
 : QTextEdit(parent, name), fInputMode(None), fRestoreInput(false), fRestorePar(0),
   fRestoreInd(0), fChar(0), fNoScroll(false), fLastBottomPos(QTextEdit::contentsHeight()),
-  fScrollBufferSize(65536), fInHistory(false), fHistPos(0), fScrollAc(0), fNonstopMode(false)
+  fScrollBufferSize(65536), fInHistory(false), fHistPos(0), fScrollAc(0), fNonstopMode(false),
+  fContextMenu(this, "game window popup menu")
 {
 	// Disable the QTextEdit's undo/redo facility, since we don't
 	// need it.  We save some memory and increase speed.
@@ -51,10 +51,18 @@ QTadsGameWindow::QTadsGameWindow( QWidget* parent, const char* name )
 	// Show the vertical scroll bar, even if it's not needed.  The
 	// user can disable it at runtime.
 	QTextEdit::setVScrollBarMode(QScrollView::AlwaysOn);
+
 	// *Always* wrap the text, no matter how (but wrapping at word
 	// boundaries will be preferred).
+	//
+	// TODO: Send a bug-report to Trolltech: setting
+	// AtWordOrDocumentBoundary works when running QTads with the
+	// same Qt version it has been compiled with, but doesn't work
+	// when running it using a different version.
+	//
 	QTextEdit::setWordWrap(QTextEdit::WidgetWidth);
 	QTextEdit::setWrapPolicy(QTextEdit::AtWordOrDocumentBoundary);
+
 	// Begin in editor-mode.
 	QTextEdit::setReadOnly(false);
 	// We only use plain text mode and format the text using
@@ -209,7 +217,7 @@ QTadsGameWindow::inputKeyPressEvent( QKeyEvent* e )
 		this->moveCursor(QTextEdit::MoveEnd, false);
 		this->doKeyboardAction(QTextEdit::ActionReturn);
 
-		{ // We use a block since we declare some temp vars.
+		{
 			// "-2" because we did a newline.
 			int par = this->paragraphs() - 2;
 			int len = this->paragraphLength(par);
@@ -273,6 +281,26 @@ QTadsGameWindow::waitCharKeyPressEvent( QKeyEvent* e )
 {
 	Q_ASSERT(this->fInputMode == WaitCharInput or this->fInputMode == PagePause);
 
+	switch (e->key()) {
+	  case Qt::Key_PageUp:
+		if (e->state() & Qt::ControlButton) {
+			this->verticalScrollBar()->subtractLine();
+		} else {
+			this->verticalScrollBar()->subtractPage();
+		}
+		e->accept();
+		return;
+
+	  case Qt::Key_PageDown:
+		if (e->state() & Qt::ControlButton) {
+			this->verticalScrollBar()->addLine();
+		} else {
+			this->verticalScrollBar()->addPage();
+		}
+		e->accept();
+		return;
+	}
+
 	if (not e->text().isEmpty()) {
 		// A key that has a Unicode value has been pressed;
 		// that's what we were waiting for.
@@ -295,7 +323,7 @@ QTadsGameWindow::charKeyPressEvent( QKeyEvent* e )
 		Q_ASSERT(this->fChar == 0);
 
 		// The character is inside the (8-bit) ASCII-range.
-		this->fChar = new QKeyEvent(*e);
+		this->fChar = new QTadsKeyEvent(*e);
 		this->fInputMode = None;
 		return;
 	}
@@ -312,7 +340,7 @@ QTadsGameWindow::charKeyPressEvent( QKeyEvent* e )
 	  case Qt::Key_F9: case Qt::Key_F10: case Qt::Key_Tab:
 		Q_ASSERT(this->fChar == 0);
 
-		this->fChar = new QKeyEvent(*e);
+		this->fChar = new QTadsKeyEvent(*e);
 		this->fInputMode = None;
 		return;
 	}
@@ -321,16 +349,6 @@ QTadsGameWindow::charKeyPressEvent( QKeyEvent* e )
 }
 
 
-void
-QTadsGameWindow::readOnlyMoveCursor( CursorAction action )
-{
-	Q_ASSERT(this->isReadOnly());
-
-	// TODO: Implement it.
-	QTextEdit::moveCursor(action, false);
-}
-
-
 void
 QTadsGameWindow::pagePause()
 {
@@ -376,6 +394,23 @@ QTadsGameWindow::keyPressEvent( QKeyEvent* e )
 		return;
 	}
 
+	// We need special handling when in read-only mode.
+	if (this->isReadOnly()) {
+		bool ctrl = e->state() & Qt::ControlButton;
+		switch (e->key()) {
+		  case Qt::Key_PageUp:
+			ctrl ? this->verticalScrollBar()->subtractLine()
+			     : this->verticalScrollBar()->subtractPage();
+			e->accept();
+			return;
+		  case Qt::Key_PageDown:
+			ctrl ? this->verticalScrollBar()->addLine()
+			     : this->verticalScrollBar()->addPage();
+			e->accept();
+			return;
+		}
+	}
+
 	if (this->fInputMode == NormalInput) {
 		// We are in normal input-mode. Call the input-mode
 		// event handler.
@@ -405,10 +440,30 @@ QTadsGameWindow::keyPressEvent( QKeyEvent* e )
 QPopupMenu*
 QTadsGameWindow::createPopupMenu( const QPoint& /* pos */ )
 {
-	QPopupMenu* menu = new QPopupMenu(this, "game window popup menu");
-	menu->insertItem(tr("Show/hide &menu"), this, SIGNAL(showHideMenu()));
-	menu->insertItem(tr("Show/hide &scrollbar"), this, SLOT(showHideScrollBar()));
-	return menu;
+	this->contextMenuEvent(0);
+	return 0;
+}
+
+
+void
+QTadsGameWindow::contextMenuEvent( QContextMenuEvent* e )
+{
+	this->fContextMenu.exec(QCursor::pos());
+	if (e != 0) e->accept();
+}
+
+
+void
+QTadsGameWindow::contentsMousePressEvent( QMouseEvent* e )
+{
+	// Although a button press is not a "full" click, we handle it
+	// as one.
+	int para;
+	int index = charAt(e->pos(), &para);
+	this->clickHandler( para, index );
+
+	// Process it.
+	QTextEdit::contentsMousePressEvent(e);
 }
 
 
@@ -417,24 +472,20 @@ QTadsGameWindow::contentsMouseMoveEvent( QMouseEvent* e )
 {
 	this->showMouseCursor();
 
+	// In case the user first pressed a mouse button before moving
+	// the mouse.  If the button gets released after the mouse has
+	// been moved out of the input-area, we must handle it as a
+	// click.
 	if (e->state() & Qt::MouseButtonMask) {
 		int para;
 		int index = charAt(e->pos(), &para);
-		clickHandler( para, index );
+		this->clickHandler( para, index );
 	}
 
 	QTextEdit::contentsMouseMoveEvent(e);
 }
 
 
-void
-QTadsGameWindow::resizeEvent( QResizeEvent* e )
-{
-	QTextEdit::resizeEvent(e);
-	// TODO: Implement it.
-}
-
-
 void
 QTadsGameWindow::focusOutEvent( QFocusEvent* )
 {
@@ -498,8 +549,8 @@ QTadsGameWindow::waitChar()
 }
 
 
-QKeyEvent
-QTadsGameWindow::getChar()
+QTadsKeyEvent
+QTadsGameWindow::getChar( bool useTimeout, int timeout )
 {
 	Q_ASSERT(this->fInputMode == None);
 
@@ -514,16 +565,37 @@ QTadsGameWindow::getChar()
 	this->getCursorPosition(&this->fInputBeginPar, &this->fInputBeginInd);
 	this->fRestorePar = this->fInputBeginPar;
 	this->fRestoreInd = this->fInputBeginInd;
-	while (this->fInputMode == RawCharInput and QTadsIO::gameRunning) {
-		qApp->eventLoop()->processEvents(QEventLoop::WaitForMore
-						 | QEventLoop::AllEvents);
+
+	if (useTimeout) {
+		// A timeout was specified; wait for an input only for
+		// the specified amount of time.
+		const QTime& curTime = QTime::currentTime();
+		while (this->fInputMode == RawCharInput and QTadsIO::gameRunning
+		       and QTime::currentTime().msecsTo(curTime) >= -timeout)
+		{
+			qApp->eventLoop()->processEvents(QEventLoop::AllEvents);
+		}
+	} else {
+		// No timeout; simply wait until an input is available.
+		while (this->fInputMode == RawCharInput and QTadsIO::gameRunning) {
+			qApp->eventLoop()->processEvents(QEventLoop::WaitForMore
+							 | QEventLoop::AllEvents);
+		}
 	}
+
 	this->setKeyCompression(true);
 	this->fLastBottomPos = this->contentsHeight();
 
 	if (not QTadsIO::gameRunning) {
 		this->fInputMode = None;
-		return QKeyEvent(QEvent::KeyPress, 0, 0, 0);
+		return QTadsKeyEvent(QEvent::KeyPress, 0, 0, 0);
+	}
+
+	if (this->fChar == 0) {
+		// No input is available; this means that the operation
+		// timed out.  Return a "timed out" QTadsKeyEvent.
+		this->fInputMode = None;
+		return QTadsKeyEvent(true);
 	}
 	return *(this->fChar);
 }
@@ -568,6 +640,11 @@ QTadsGameWindow::insert( std::queue<QTadsFormattedString>& text, uint )
 		if (text.front().f.italics != this->italic()) {
 			this->setItalic(text.front().f.italics);
 		}
+		if (not text.front().f.color.isValid()) {
+			this->setColor(this->paletteForegroundColor());
+		} else if (text.front().f.color != this->color()) {
+			this->setColor(text.front().f.color);
+		}
 		// Display the text.
 		QTextEdit::insert(text.front().s);
 		// Remove the text from the buffer.
@@ -592,16 +669,30 @@ QTadsGameWindow::insert( std::queue<QTadsFormattedString>& text, uint )
 			QTadsIO::clearSysStatus();
 			this->fLastBottomPos = this->contentsY() + this->visibleHeight();
 		}
-		// If a game is currently running, scroll softly.  If
-		// not, simple scroll to the bottom.
+		// If a game is currently running, scroll softly (if
+		// enabled in the settings).  If not, simple scroll to
+		// the bottom.
 		if (QTadsIO::gameRunning) {
-			this->verticalScrollBar()->addLine();
+			if (QTadsIO::settings().softScrolling()) {
+				// Scroll by a line.
+				this->verticalScrollBar()->addLine();
+			} else {
+				// Scroll immediately to the next
+				// "more" point.
+				this->verticalScrollBar()
+					->setValue(this->verticalScrollBar()->value()
+						   + this->fLastBottomPos
+						   - this->contentsY()
+						   - this->verticalScrollBar()->lineStep());
+			}
 		} else {
+			// Scroll to the bottom.
 			this->verticalScrollBar()
 				->setValue(this->verticalScrollBar()->maxValue());
 		}
 		// Process any drawing.
 		qApp->eventLoop()->processEvents(QEventLoop::AllEvents, 1);
+		this->repaintChanged();
 	}
 
 	// If the text grew too large, delete an amount of the oldest
@@ -642,11 +733,6 @@ QTadsGameWindow::setContentsPos( int x, int y )
 void
 QTadsGameWindow::moveCursor( CursorAction action, bool )
 {
-	//if (this->isReadOnly()) {
-	//	this->readOnlyMoveCursor(action);
-	//	return;
-	//}
-
 	// We need to enable immediate widget-updates when moving the
 	// cursor, or else the user will not be able to see what he's
 	// doing.  If updates are currently disabled, enable them and
@@ -843,6 +929,22 @@ QTadsGameWindow::paste()
 }
 
 
+void
+QTadsGameWindow::contentsDropEvent( QDropEvent* evnt )
+{
+	if (this->fInputMode == NormalInput
+	    and this->paragraphAt(evnt->pos()) == this->fInputBeginPar
+	    and this->charAt(evnt->pos(), 0) >= this->fInputBeginInd)
+	{
+		// It's within the editable area; allow the operation.
+		QTextEdit::contentsDropEvent(evnt);
+	} else {
+		// It's non in the input-are; ignore it.
+		evnt->ignore();
+	}
+}
+
+
 void
 QTadsGameWindow::enterCommand( const QString& cmd )
 {
@@ -885,7 +987,7 @@ QTadsGameWindow::clickHandler( int para, int pos )
 		this->setReadOnly(true);
 		this->updateContents(this->paragraphRect(para));
 		this->fRestoreInput = true;
-	} else if (indexTo != -1 &&
+	} else if (indexTo != -1 and
 		   (indexFrom < this->fInputBeginInd or paraFrom < this->fInputBeginPar)) {
 		// There is a selection extending outside the
 		// input-area; we can't disable read-only mode, since
