@@ -277,7 +277,8 @@ static errmdef errlist[] =
     { ERR_INVOPC, "invalid opcode (internal error)" },
     { ERR_RUNNOBJ,"property evaluated for non-existent object" },
     { ERR_EXTLOAD,"unable to load external function \"%s\"" },
-    { ERR_EXTRUN, "error executing external function \"%s\"" },
+    { ERR_EXTRUN,
+    "external functions are obsolete (attempted to invoke \"%s\")" },
     { ERR_CIRCSYN,"circular synonym" },
     { ERR_DIVZERO,"divide by zero" },
     { ERR_BADDEL, "only objects allocated with 'new' can be deleted" },
@@ -442,10 +443,12 @@ use with your version of the debugger)" },
     /* tr security suboptions help */
     { ERR_TRSUS1,     "-s (file I/O safety level) suboptions:" },
     { ERR_TRSUS1 + 1,
-       "  -s0    (default) minimum safety - read and write in any directory" },
+    "  -s0    minimum safety - read and write in any directory" },
     { ERR_TRSUS1 + 2,
-       "  -s1    read in any directory, write in current directory" },
-    { ERR_TRSUS1 + 3, "  -s2    read-only access in any directory" },
+    "  -s1    read in any directory, write in current directory only"
+    },
+    { ERR_TRSUS1 + 3,
+    "  -s2    (default) read and write in current directory only" },
     { ERR_TRSUS1 + 4, "  -s3    read-only access in current directory only" },
     { ERR_TRSUS1 + 5, "  -s4    maximum safety - no file I/O allowed" },
     { ERR_TRSUS1 + 6,
