@@ -240,7 +240,6 @@ static void fiord1(mcmcxdef *mctx, voccxdef *vctx, tokcxdef *tctx,
                    appctxdef *appctx, char *argv0)
 {
     int         i;
-    int         j;
     int         siz;
     uchar       buf[TOKNAMMAX + 50];
     errcxdef   *ec = vctx->voccxerr;
@@ -491,6 +490,11 @@ static void fiord1(mcmcxdef *mctx, voccxdef *vctx, tokcxdef *tctx,
                 errsig(ec, ERR_RDGAM);
             siz = osrp2(buf);
 
+#if 0
+/* 
+ *   external functions are now obsolete - do not load
+ */            
+
             /* look for an external function with the same name */
             for (i = vctx->voccxrun->runcxexc,  ex = vctx->voccxrun->runcxext
                  ; i ; ++ex, --i)
@@ -503,18 +507,18 @@ static void fiord1(mcmcxdef *mctx, voccxdef *vctx, tokcxdef *tctx,
             /* if we found an external function of this name, load it */
             if (i && !xfcns_done)
             {
+                /* load the function */
                 ex->runxptr = os_exfld(fp, (unsigned)siz);
-#ifdef NEVER
-                ex->runxptr = (int (*)())mchalo(ec, (ushort)siz, "XFCN");
-                if (osfrb(fp, (uchar *)ex->runxptr, (uint)siz))
-                    errsig(ec, ERR_RDGAM);
-#endif
             }
             else
             {
                 /* this XFCN isn't used; don't bother loading it */
                 osfseek(fp, endpos + startofs, OSFSK_SET);
             }
+#else
+            /* external functions are obsolete; simply skip the data */
+            osfseek(fp, endpos + startofs, OSFSK_SET);
+#endif
         }
         else if (fioisrsc(buf, "HTMLRES"))
         {
@@ -977,6 +981,22 @@ void fiord(mcmcxdef *mctx, voccxdef *vctx, tokcxdef *tctx, char *fname,
      *   provided, otherwise use the name of the executable file itself 
      */
     display_fname = (fname != 0 ? fname : exename);
+
+    /* save the filename in G_os_gamename */
+    if (display_fname != 0)
+    {
+        size_t copylen;
+
+        /* limit the copy to the buffer size */
+        if ((copylen = strlen(display_fname)) > sizeof(G_os_gamename) - 1)
+            copylen = sizeof(G_os_gamename) - 1;
+
+        /* save it */
+        memcpy(G_os_gamename, display_fname, copylen);
+        G_os_gamename[copylen] = '\0';
+    }
+    else
+        G_os_gamename[0] = '\0';
     
     /* open the file and read and check file header */
     fp = (fname != 0 ? osfoprb(fname, OSFTGAME)
