@@ -46,6 +46,8 @@ Modified
 #include "vmfunc.h"
 #include "vmpat.h"
 #include "vmtobj.h"
+#include "vmvec.h"
+#include "vmpredef.h"
 
 
 /* ------------------------------------------------------------------------ */
@@ -596,6 +598,8 @@ void CVmBifTADS::rand(VMG_ uint argc)
     int choose_an_arg;
     const char *listp;
     ulong rand_val;
+    CVmObjVector *vec = 0;
+    vm_obj_id_t vecid = VM_INVALID_OBJ;
 
     /* presume we're not going to choose from our arguments or from a list */
     choose_an_arg = FALSE;
@@ -621,12 +625,30 @@ void CVmBifTADS::rand(VMG_ uint argc)
     }
     else if (argc == 1)
     {
-        /* pop the list value */
-        listp = pop_list_val(vmg0_);
+        /* check for a vector or a list */
+        if (G_stk->get(0)->typ == VM_OBJ
+            && CVmObjVector::is_vector_obj(vmg_ G_stk->get(0)->val.obj))
+        {
+            /* 
+             *   it's a vector - get the object pointer, but leave it on the
+             *   stack for GC protection for now 
+             */
+            vecid = G_stk->get(0)->val.obj;
+            vec = (CVmObjVector *)vm_objp(vmg_ vecid);
 
-        /* our range is 0..(list_element_count-1) */
-        range = vmb_get_len(listp);
-        use_range = TRUE;
+            /* the range is 0..(vector_length-1) */
+            range = vec->get_element_count();
+            use_range = TRUE;
+        }
+        else
+        {
+            /* it must be a list - pop the list value */
+            listp = pop_list_val(vmg0_);
+
+            /* our range is 0..(list_element_count-1) */
+            range = vmb_get_len(listp);
+            use_range = TRUE;
+        }
     }
     else
     {
@@ -697,6 +719,31 @@ void CVmBifTADS::rand(VMG_ uint argc)
         /* discard all of the arguments */
         G_stk->discard(argc);
     }
+    else if (vec != 0)
+    {
+        vm_val_t val;
+
+        /* get the selected element */
+        if (range == 0)
+        {
+            /* there are no elements to choose from, so return nil */
+            val.set_nil();
+        }
+        else
+        {
+            vm_val_t idxval;
+
+            /* get the selected vector element */
+            idxval.set_int(rand_val + 1);
+            vec->index_val(vmg_ &val, vecid, &idxval);
+        }
+
+        /* set the result */
+        retval(vmg_ &val);
+
+        /* discard our gc protection */
+        G_stk->discard();
+    }
     else if (listp != 0)
     {
         vm_val_t val;
@@ -841,7 +888,20 @@ void CVmBifTADS::cvtnum(VMG_ uint argc)
         return;
     }
 
-    /* we can only convert a string */
+    /* if it's already an integer, just return the same value */
+    if (valp->typ == VM_INT)
+    {
+        /* just return the argument value */
+        retval_int(vmg_ valp->val.intval);
+
+        /* discard arguments (ignore the radix in this case) */
+        G_stk->discard(argc);
+
+        /* done */
+        return;
+    }
+
+    /* otherwise, it must be a string */
     strp = pop_str_val(vmg0_);
     len = vmb_get_len(strp);
 
@@ -854,6 +914,7 @@ void CVmBifTADS::cvtnum(VMG_ uint argc)
         /* make sure the radix is valid */
         switch(radix)
         {
+        case 2:
         case 8:
         case 10:
         case 16:
@@ -918,6 +979,16 @@ void CVmBifTADS::cvtnum(VMG_ uint argc)
         /* scan the digits */
         switch (radix)
         {
+        case 2:
+            for ( ; rem != 0 && (p.getch() == '0' || p.getch() == '1') ;
+                  p.inc(&rem))
+            {
+                acc <<= 1;
+                if (p.getch() == '1')
+                    acc += 1;
+            }
+            break;
+            
         case 8:
             for ( ; rem != 0 && is_odigit(p.getch()) ; p.inc(&rem))
             {
@@ -1427,8 +1498,7 @@ void CVmBifTADS::re_group(VMG_ uint argc)
      *   get the previous search string - get a pointer directly to the
      *   contents of the string
      */
-    last_str = G_bif_tads_globals->last_rex_str->val.get_as_string(vmg0_)
-               + VMB_LEN;
+    last_str = G_bif_tads_globals->last_rex_str->val.get_as_string(vmg0_);
 
     /* get the register */
     reg = G_bif_tads_globals->rex_searcher->get_group_reg(groupno);
@@ -1461,7 +1531,7 @@ void CVmBifTADS::re_group(VMG_ uint argc)
      *   the search, not the start of the string, so we need to add in the
      *   starting point in the search.  
      */
-    p.set((char *)last_str);
+    p.set((char *)last_str + VMB_LEN);
     put_list_int(&dst, p.len(start_byte_ofs) + 1);
 
     /* 
@@ -2285,8 +2355,27 @@ void CVmBifTADS::get_func_params(VMG_ uint argc)
     /* check arguments */
     check_argc(vmg_ argc, 1);
 
-    /* get the function pointer */
-    G_interpreter->pop_funcptr(vmg_ &func);
+    /* the argument can be an anonymous function object or function pointer */
+    if (G_stk->get(0)->typ == VM_OBJ
+        && G_predef->obj_call_prop != VM_INVALID_PROP)
+    {
+        uint argc = 0;
+        vm_obj_id_t srcobj;
+        
+        /* it's an anonymous function - get the object */
+        G_interpreter->pop_obj(vmg_ &func);
+
+        /* retrieve its ObjectCallProp value, and make sure it's a function */
+        if (!vm_objp(vmg_ func.val.obj)->get_prop(
+            vmg_ G_predef->obj_call_prop, &func, func.val.obj, &srcobj, &argc)
+            || func.typ != VM_FUNCPTR)
+            err_throw(VMERR_FUNCPTR_VAL_REQD);
+    }
+    else
+    {
+        /* it's a simple function pointer - retrieve it */
+        G_interpreter->pop_funcptr(vmg_ &func);
+    }
 
     /* set up a pointer to the function header */
     hdr.set((const uchar *)G_code_pool->get_ptr(func.val.ofs));
