@@ -139,6 +139,7 @@ void CVmBifTIO::say_to_console(VMG_ CVmConsole *console, uint argc)
 
 #define LOG_SCRIPT   1
 #define LOG_CMD      2
+#define LOG_EVENT    3
 
 void CVmBifTIO::logging(VMG_ uint argc)
 {
@@ -168,6 +169,7 @@ void CVmBifTIO::logging(VMG_ uint argc)
             break;
 
         case LOG_CMD:
+        case LOG_EVENT:
             G_console->close_command_log();
             break;
 
@@ -194,7 +196,8 @@ void CVmBifTIO::logging(VMG_ uint argc)
             break;
 
         case LOG_CMD:
-            G_console->open_command_log(fname);
+        case LOG_EVENT:
+            G_console->open_command_log(fname, log_type == LOG_EVENT);
             break;
 
         default:
@@ -271,34 +274,51 @@ void CVmBifTIO::input(VMG_ uint argc)
 void CVmBifTIO::inputkey(VMG_ uint argc)
 {
     char buf[32];
-    char c[4];
+    char c[10];
     size_t len;
+    int evt;
+    static const int filter[] = { OS_EVT_KEY };
 
     /* check arguments */
     check_argc(vmg_ argc, 0);
 
+    /* check for script input */
+    if (G_console->read_event_script(
+        vmg_ &evt, buf, sizeof(buf),
+        filter, sizeof(filter)/sizeof(filter[0]), 0))
+    {
+        /* we got a key from the script */
+        retval_ui_str(vmg_ buf);
+
+        /* log the event */
+        G_console->log_event(vmg_ OS_EVT_KEY, buf, strlen(buf), FALSE);
+
+        /* done */
+        return;
+    }
+
     /* flush any output */
     G_console->flush_all(vmg_ VM_NL_INPUT);
-
+    
     /* get a keystroke */
-    c[0] = os_getc_raw();
+    c[0] = (char)os_getc_raw();
     len = 1;
-
+    
     /* if it's an extended key, map it specially */
     if (c[0] == 0)
     {
         char extc;
-
+        
         /* get the second part of the sequence */
-        extc = os_getc_raw();
-
+        extc = (char)os_getc_raw();
+        
         /* map the extended key */
         map_ext_key(vmg_ buf, (unsigned char)extc);
     }
     else
     {
         /* continue fetching bytes until we have a full character */
-        while (!raw_key_complete(vmg_ c, len) && len < sizeof(c))
+        while (!raw_key_complete(vmg_ c, len) && len + 1 < sizeof(c))
         {
             /* 
              *   We don't yet have enough bytes for a complete character, so
@@ -308,19 +328,23 @@ void CVmBifTIO::inputkey(VMG_ uint argc)
              *   os_getc_raw() here - it should simply return the next byte
              *   of the sequence immediately.  
              */
-            c[len++] = os_getc_raw();
+            c[len++] = (char)os_getc_raw();
         }
-
+        c[len] = '\0';
+        
         /* 
-         *   translate the key from the local character set to UTF-8, and
-         *   map the extended key code to the portable representation 
+         *   translate the key from the local character set to UTF-8, and map
+         *   the extended key code to the portable representation 
          */
         map_raw_key(vmg_ buf, c, len);
     }
-
+    
     /* reset the [MORE] counter */
     G_console->reset_line_count(FALSE);
-
+    
+    /* log the event */
+    G_console->log_event(vmg_ OS_EVT_KEY, buf, strlen(buf), TRUE);
+    
     /* return the string */
     retval_str(vmg_ buf);
 }
@@ -331,7 +355,6 @@ void CVmBifTIO::inputkey(VMG_ uint argc)
  */
 void CVmBifTIO::inputevent(VMG_ uint argc)
 {
-    char keyname[32];
     int use_timeout;
     unsigned long timeout;
     os_event_info_t info;
@@ -339,7 +362,14 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
     int ele_count;
     vm_obj_id_t lst_obj;
     CVmObjList *lst;
+    char keyname[32];
     vm_val_t val;
+    int from_script = FALSE;
+    static const int filter[] =
+    {
+        OS_EVT_KEY, OS_EVT_TIMEOUT, OS_EVT_NOTIMEOUT,
+        OS_EVT_HREF, OS_EVT_EOF, OS_EVT_COMMAND
+    };
 
     /* check arguments */
     check_argc_range(vmg_ argc, 0, 1);
@@ -369,30 +399,55 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
         use_timeout = TRUE;
     }
 
-    /* flush any buffered output */
-    G_console->flush_all(vmg_ VM_NL_INPUT);
+    /* check for script input */
+    if (G_console->read_event_script(
+        vmg_ &evt, info.href, sizeof(info.href),
+        filter, sizeof(filter)/sizeof(filter[0]), 0))
+    {
+        /* we got a script event - note it */
+        from_script = TRUE;
 
-    /* reset the [MORE] counter */
-    G_console->reset_line_count(FALSE);
+        /* translate certain events */
+        switch (evt)
+        {
+        case OS_EVT_COMMAND:
+            /* read the numeric parameter */
+            info.cmd_id = atoi(info.href);
+            break;
+        }
+    }
+    else
+    {
+        /* flush any buffered output */
+        G_console->flush_all(vmg_ VM_NL_INPUT);
 
-    /* read an event */
-    evt = os_get_event(timeout, use_timeout, &info);
+        /* reset the [MORE] counter */
+        G_console->reset_line_count(FALSE);
+
+        /* read an event from the OS layer */
+        evt = os_get_event(timeout, use_timeout, &info);
+    }
 
     /* figure out how big a list we need to allocate */
     switch(evt)
     {
     case OS_EVT_KEY:
         /* 
-         *   we need two elements - one for the event type code, one for
-         *   the keystroke string 
+         *   we need two elements - one for the event type code, one for the
+         *   keystroke string 
          */
         ele_count = 2;
         break;
 
+    case OS_EVT_COMMAND:
+        /* we need a second element for the command ID */
+        ele_count = 2;
+        break;
+
     case OS_EVT_HREF:
         /* 
-         *   we need two elements - one for the event type code, one for
-         *   the HREF string 
+         *   we need two elements - one for the event type code, one for the
+         *   HREF string 
          */
         ele_count = 2;
         break;
@@ -420,10 +475,27 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
     {
     case OS_EVT_KEY:
         /* map the extended or ordinary key, as appropriate */
-        if (info.key[0] == 0)
+        if (from_script)
+        {
+            /* we got a key from the script - it's in the 'href' field */
+            val.set_obj(str_from_ui_str(vmg_ info.href));
+
+            /* log the event */
+            G_console->log_event(
+                vmg_ OS_EVT_KEY, info.href, strlen(info.href), FALSE);
+        }
+        else if (info.key[0] == 0)
         {
             /* it's an extended key */
             map_ext_key(vmg_ keyname, info.key[1]);
+
+            /* create a string for the key name */
+            val.set_obj(CVmObjString::create(
+                vmg_ FALSE, keyname, strlen(keyname)));
+
+            /* log the event */
+            G_console->log_event(
+                vmg_ OS_EVT_KEY, keyname, strlen(keyname), TRUE);
         }
         else
         {
@@ -431,7 +503,7 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
             size_t len;
 
             /* fetch more bytes until we have a complete character */
-            for (c[0] = info.key[0], len = 1 ;
+            for (c[0] = (char)info.key[0], len = 1 ;
                  !raw_key_complete(vmg_ c, len) && len < sizeof(c) ; )
             {
                 /* 
@@ -453,32 +525,55 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
                     break;
 
                 /* store the next byte of the sequence */
-                c[len++] = info.key[0];
+                c[len++] = (char)info.key[0];
             }
 
             /* it's an ordinary key - map it */
             map_raw_key(vmg_ keyname, c, len);
+
+            /* create a string for the key name */
+            val.set_obj(CVmObjString::create(
+                vmg_ FALSE, keyname, strlen(keyname)));
+
+            /* log the event */
+            G_console->log_event(
+                vmg_ OS_EVT_KEY, keyname, strlen(keyname), TRUE);
         }
-        
-        /* create a string for the key name */
-        val.set_obj(CVmObjString::create(vmg_ FALSE, keyname,
-                                         strlen(keyname)));
 
         /* add it to the list */
         lst->cons_set_element(1, &val);
+
         break;
 
     case OS_EVT_HREF:
         /* create the string for the href text */
-        val.set_obj(CVmObjString::create(vmg_ FALSE, info.href,
-                                         strlen(info.href)));
+        val.set_obj(str_from_ui_str(vmg_ info.href));
 
         /* add it to the list */
         lst->cons_set_element(1, &val);
+
+        /* log it */
+        G_console->log_event(vmg_ OS_EVT_HREF,
+                             info.href, strlen(info.href), FALSE);
+        break;
+
+    case OS_EVT_COMMAND:
+        /* the second element is the command ID code */
+        val.set_int(info.cmd_id);
+        lst->cons_set_element(1, &val);
+
+        /* log it */
+        {
+            char buf[20];
+            sprintf(buf, "%d", info.cmd_id);
+            G_console->log_event(vmg_ OS_EVT_COMMAND,
+                                 buf, strlen(buf), TRUE);
+        }
         break;
 
     default:
         /* other event types have no extra data */
+        G_console->log_event(vmg_ evt);
         break;
     }
 
@@ -489,7 +584,6 @@ void CVmBifTIO::inputevent(VMG_ uint argc)
     G_stk->discard();
 }
 
-
 /* ------------------------------------------------------------------------ */
 /*
  *   Service routine: Map an "extended" keystroke from raw os_getc_raw()
@@ -537,6 +631,7 @@ int CVmBifTIO::map_ext_key(VMG_ char *namebuf, int extc)
         "[word-right]",                              /* CMD_WORD_RIGHT - 30 */
         "[del-word]",                                  /* CMD_WORDKILL - 31 */
         "[eof]",                                            /* CMD_EOF - 32 */
+        "[break]"                                         /* CMD_BREAK - 33 */
     };
 
     /* if it's in the key name array, use the array entry */
@@ -663,6 +758,7 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
     int default_resp;
     int cancel_resp;
     int resp;
+    char numbuf[32];
     
     /* check arguments */
     check_argc(vmg_ argc, 5);
@@ -671,7 +767,7 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
     icon_id = pop_int_val(vmg0_);
 
     /* get the prompt string */
-    pop_str_val_buf(vmg_ prompt, sizeof(prompt));
+    pop_str_val_ui(vmg_ prompt, sizeof(prompt));
 
     /* there aren't any buttons yet */
     btn_cnt = 0;
@@ -839,13 +935,35 @@ void CVmBifTIO::inputdialog(VMG_ uint argc)
         cancel_resp = pop_int_val(vmg0_);
     }
 
-    /* flush output before showing the dialog */
-    G_console->flush_all(vmg_ VM_NL_INPUT);
+    /* check for script input */
+    static int filter[] = { VMCON_EVT_DIALOG };
+    int evt;
+    if (G_console->read_event_script(
+        vmg_ &evt, numbuf, sizeof(numbuf),
+        filter, sizeof(filter)/sizeof(filter[0]), 0))
+    {
+        /* we got a script response - no need to show the dialog */
+        resp = atoi(numbuf);
 
-    /* show the dialog */
-    resp = G_console->input_dialog(vmg_ icon_id, prompt,
-                                   std_btns, labels, btn_cnt,
-                                   default_resp, cancel_resp);
+        /* log the event */
+        G_console->log_event(vmg_ VMCON_EVT_DIALOG,
+                             numbuf, strlen(numbuf), FALSE);
+    }
+    else
+    {
+        /* flush output before showing the dialog */
+        G_console->flush_all(vmg_ VM_NL_INPUT);
+
+        /* show the dialog */
+        resp = G_console->input_dialog(vmg_ icon_id, prompt,
+                                       std_btns, labels, btn_cnt,
+                                       default_resp, cancel_resp);
+
+        /* log the event */
+        sprintf(numbuf, "%d", resp);
+        G_console->log_event(vmg_ VMCON_EVT_DIALOG,
+                             numbuf, strlen(numbuf), TRUE);
+    }
 
     /* return the result */
     retval_int(vmg_ resp);
@@ -866,12 +984,13 @@ void CVmBifTIO::askfile(VMG_ uint argc)
     vm_obj_id_t lst_obj;
     CVmObjList *lst;
     vm_val_t val;
+    int from_script = FALSE;
     
     /* check arguments */
     check_argc(vmg_ argc, 4);
 
     /* get the prompt string */
-    pop_str_val_buf(vmg_ prompt, sizeof(prompt));
+    pop_str_val_ui(vmg_ prompt, sizeof(prompt));
 
     /* get the dialog type and file type */
     dialog_type = pop_int_val(vmg0_);
@@ -880,9 +999,135 @@ void CVmBifTIO::askfile(VMG_ uint argc)
     /* pop the flags */
     flags = pop_long_val(vmg0_);
 
-    /* ask for a file */
-    result = G_console->askfile(vmg_ prompt, strlen(prompt),
-                                fname, sizeof(fname), dialog_type, file_type);
+    /* check for a script response */
+    static int filter[] = { VMCON_EVT_FILE };
+    int evt;
+    unsigned long attrs;
+    if (G_console->read_event_script(
+        vmg_ &evt, fname, sizeof(fname),
+        filter, sizeof(filter)/sizeof(filter[0]), &attrs))
+    {
+        char prompt[OSFNMAX + 255];
+        int ok = TRUE;
+        
+        /* we got a response from the script */
+        from_script = TRUE;
+        result = (fname[0] != '\0' ? OS_AFE_SUCCESS : OS_AFE_CANCEL);
+
+        /* 
+         *   If this is a "save" prompt, and the OVERWRITE flag isn't
+         *   provided, and the file already exists, show an interactive
+         *   warning that we're about to ovewrite the file.  
+         */
+        if (dialog_type == OS_AFP_SAVE
+            && result == OS_AFE_SUCCESS
+            && (attrs & VMCON_EVTATTR_OVERWRITE) == 0
+            && !osfacc(fname))
+        {
+            /* the file exists - warn about the overwrite */
+            ok = FALSE;
+            sprintf(prompt, "The script might overwrite the file %s. ", fname);
+        }
+
+        /* 
+         *   If this is a "save" prompt, AND the file doesn't already exist,
+         *   try creating the file to see if we can - this will test to see
+         *   if the target directory exists and is writable, for instance.  
+         */
+        if (ok
+            && dialog_type == OS_AFP_SAVE
+            && result == OS_AFE_SUCCESS
+            && osfacc(fname))
+        {
+            /* try creating the file */
+            osfildef *fp = osfopwb(fname, file_type);
+
+            /* if that succeeded, undo the creation; otherwise warn */
+            if (fp != 0)
+            {
+                /* it worked - close and delete the test file */
+                osfcls(fp);
+                osfdel(fname);
+            }
+            else
+            {
+                /* didn't work - warn about it */
+                ok = FALSE;
+                sprintf(prompt, "The script is attempting to create file %s, "
+                        "but that file cannot be created.", fname);
+            }
+        }
+
+        /*
+         *   If this is an "open" prompt, and the file isn't readable, warn
+         *   about it. 
+         */
+        if (ok
+            && dialog_type == OS_AFP_OPEN
+            && result == OS_AFE_SUCCESS
+            && osfacc(fname))
+        {
+            ok = FALSE;
+            sprintf(prompt, "The script is attempting to open file %s, "
+                    "but this file doesn't exist or isn't readable.", fname);
+        }
+
+        /* check to see if we're warning about anything */
+        if (!ok)
+        {
+            char fullprompt[OSFNMAX + 255 + 150];
+
+            /* build the full prompt */
+            sprintf(fullprompt, "%s Do you wish to proceed? Select Yes to "
+                    "proceed with this file, No to choose a different file, "
+                    "or Cancel to stop script playback.", prompt);
+
+            /* 
+             *   display a dialog - note that this goes directly to user,
+             *   bypassing the active script, since this is a question about
+             *   how to handle a problem in the script 
+             */
+            switch (G_console->input_dialog(
+                vmg_ OS_INDLG_ICON_WARNING, fullprompt,
+                OS_INDLG_YESNOCANCEL, 0, 0, 2, 3))
+            {
+            case 1:
+                /* yes - proceed with fname */
+                break;
+
+            case 2:
+                /* no - ask for a new file */
+                result = G_console->askfile(
+                    vmg_ prompt, strlen(prompt),
+                    fname, sizeof(fname), dialog_type, file_type);
+
+                /* this didn't come from the script after all */
+                from_script = FALSE;
+
+                /* if they canceled, cancel the whole script playback */
+                if (result != OS_AFE_SUCCESS)
+                    close_script_file(vmg0_);
+
+                /* handled */
+                break;
+
+            case 3:
+                /* cancel - stop the script playback */
+                close_script_file(vmg0_);
+
+                /* indicate cancellation */
+                result = OS_AFE_CANCEL;
+                break;
+            }
+        }
+    }
+    else
+    {
+        /* ask for a file via the UI */
+        result = G_console->askfile(
+            vmg_ prompt, strlen(prompt),
+            fname, sizeof(fname), dialog_type, file_type);
+    }
 
     /* 
      *   Allocate a list to store the return value.  If we successfully
@@ -906,13 +1151,27 @@ void CVmBifTIO::askfile(VMG_ uint argc)
     /* if we got a filename, set the second element to the filename string */
     if (result == OS_AFE_SUCCESS)
     {
-        /* create a string for the filename */
-        val.set_obj(CVmObjString::create(vmg_ FALSE, fname, strlen(fname)));
+        /* 
+         *   Create a string for the filename.  If it's coming from a script,
+         *   we need to translate it to the local character set; otherwise
+         *   it's already in UTF-8. 
+         */
+        val.set_obj(
+            from_script
+            ? str_from_ui_str(vmg_ fname)
+            : CVmObjString::create(vmg_ FALSE, fname, strlen(fname)));
 
         /* store the string as the second list element */
         lst->cons_set_element(1, &val);
     }
 
+    /* log the event */
+    if (result == OS_AFE_SUCCESS)
+        G_console->log_event(vmg_ VMCON_EVT_FILE,
+                             fname, strlen(fname), FALSE);
+    else
+        G_console->log_event(vmg_ VMCON_EVT_FILE);
+
     /* return the list */
     retval_obj(vmg_ lst_obj);
 
@@ -1135,6 +1394,7 @@ void CVmBifTIO::res_exists(VMG_ uint argc)
 /* turn off 'more' mode while reading the script */
 #define VMBIFTADS_SCRIPT_NONSTOP   0x0002
 
+
 /*
  *   set_script_file - open a command input scripting file 
  */
@@ -1149,8 +1409,6 @@ void CVmBifTIO::set_script_file(VMG_ uint argc)
     /* if the filename is nil, close the current script file */
     if (G_stk->get(0)->typ == VM_NIL)
     {
-        int old_more_mode;
-        
         /* discard the nil filename */
         G_stk->discard();
         
@@ -1159,10 +1417,7 @@ void CVmBifTIO::set_script_file(VMG_ uint argc)
             G_stk->discard();
 
         /* close the script file */
-        old_more_mode = G_console->close_script_file();
-
-        /* restore the MORE mode in effect when the script was opened */
-        G_console->set_more_state(old_more_mode);
+        close_script_file(vmg0_);
     }
     else
     {
@@ -1184,6 +1439,20 @@ void CVmBifTIO::set_script_file(VMG_ uint argc)
     }
 }
 
+/*
+ *   close the script file 
+ */
+void CVmBifTIO::close_script_file(VMG0_)
+{
+    int old_more_mode;
+
+    /* close the script */
+    old_more_mode = G_console->close_script_file();
+
+    /* restore the MORE mode in effect when the script was opened */
+    G_console->set_more_state(old_more_mode);
+}
+
 /* ------------------------------------------------------------------------ */
 /*
  *   get_charset selectors 
@@ -1192,8 +1461,11 @@ void CVmBifTIO::set_script_file(VMG_ uint argc)
 /* display character set */
 #define VMBIFTADS_CHARSET_DISPLAY  0x0001
 
-/* file system character set */
-#define VMBIFTADS_CHARSET_FILE     0x0002
+/* file system character set for filenames */
+#define VMBIFTADS_CHARSET_FILENAME 0x0002
+
+/* typical character set for text file contents */
+#define VMBIFTADS_CHARSET_FILECONTENTS 0x0003
 
 /*
  *   get_charset - get a local character set name 
@@ -1230,10 +1502,14 @@ void CVmBifTIO::get_charset(VMG_ uint argc)
         which = OS_CHARMAP_DISPLAY;
         break;
 
-    case VMBIFTADS_CHARSET_FILE:
+    case VMBIFTADS_CHARSET_FILENAME:
         which = OS_CHARMAP_FILENAME;
         break;
 
+    case VMBIFTADS_CHARSET_FILECONTENTS:
+        which = OS_CHARMAP_FILECONTENTS;
+        break;
+
     default:
         /* others are unrecognized; simply return nil for these */
         retval_nil(vmg0_);
@@ -1829,12 +2105,20 @@ void CVmBifTIO::log_console_create(VMG_ uint argc)
      *   Retrieve the character mapper, which can be given as either a
      *   CharacterSet object or a string giving the character set name.  
      */
-    if (G_stk->get(0)->typ == VM_OBJ
-        && CVmObjCharSet::is_charset(vmg_ G_stk->get(0)->val.obj))
+    if (G_stk->get(0)->typ == VM_NIL)
+    {
+        /* nil - use the default log file character set */
+        cmap = G_cmap_to_log;
+
+        /* add our reference to it */
+        cmap->add_ref();
+    }
+    else if (G_stk->get(0)->typ == VM_OBJ
+             && CVmObjCharSet::is_charset(vmg_ G_stk->get(0)->val.obj))
     {
         vm_obj_id_t obj;
         
-        /* retrieve the CharacterSet reference */
+        /* it's a CharacterSet object - pop the reference */
         obj = CVmBif::pop_obj_val(vmg0_);
 
         /* retrieve the character mapper from the character set */
@@ -1952,16 +2236,33 @@ void CVmBifTIO::log_console_close(VMG_ uint argc)
  */
 void CVmBifTIO::log_console_say(VMG_ uint argc)
 {
+    int hdl;
     CVmConsole *console;
+    CVmConsoleMainLog main_log_console;
 
     /* check arguments */
     check_argc_range(vmg_ argc, 1, 32767);
 
-    /* get the banner - if it's invalid, throw an error */
-    console = G_console->get_log_console_manager()
-              ->get_console(pop_int_val(vmg0_));
-    if (console == 0)
-        err_throw(VMERR_BAD_VAL_BIF);
+    /* get the console handle */
+    hdl = pop_int_val(vmg0_);
+
+    /* 
+     *   if it's the special value -1, it means that we want to write to the
+     *   main console's log file; otherwise, it's a log console that we
+     *   previously created explicitly via log_console_create()
+     */
+    if (hdl == -1)
+    {
+        /* use the main log */
+        console = &main_log_console;
+    }
+    else
+    {
+        /* get the console by handle - if it's invalid, throw an error */
+        console = G_console->get_log_console_manager()->get_console(hdl);
+        if (console == 0)
+            err_throw(VMERR_BAD_VAL_BIF);
+    }
 
     /*
      *   write the argument(s) to the console (note that the first argument,
