@@ -87,7 +87,7 @@ int (CVmObjByteArray::
 vm_obj_id_t CVmObjByteArray::create_from_stack(VMG_ const uchar **pc_ptr,
                                                uint argc)
 {
-    vm_obj_id_t id;
+    vm_obj_id_t id = VM_INVALID_OBJ;
     CVmObjByteArray *arr;
     unsigned long cnt;
     vm_val_t *arg1;
@@ -384,7 +384,7 @@ int CVmObjByteArray::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                               vm_obj_id_t self, vm_obj_id_t *source_obj,
                               uint *argc)
 {
-    ushort func_idx;
+    uint func_idx;
 
     /* translate the property index to an index into our function table */
     func_idx = G_meta_table
@@ -623,7 +623,7 @@ void CVmObjByteArray::load_image_data(VMG_ const char *ptr, size_t siz)
     notify_delete(vmg_ FALSE);
 
     /* get the new array size */
-    cnt = osrp4(ptr);
+    cnt = t3rp4u(ptr);
 
     /* make sure the size isn't larger than we'd expect */
     if (siz > 4 + cnt)
@@ -711,7 +711,7 @@ void CVmObjByteArray::restore_from_file(VMG_ vm_obj_id_t self,
     unsigned long rem;
 
     /* read the element count */
-    ele_cnt = (unsigned long)fp->read_int4();
+    ele_cnt = fp->read_uint4();
 
     /* allocate or reallocate as needed */
     if (ext_ == 0)
@@ -906,13 +906,13 @@ uint CVmObjByteArray::calc_hash(VMG_ vm_obj_id_t self, int /*depth*/) const
         if (chunk > rem)
             chunk = (size_t)rem;
 
+        /* advance our counters for this chunk */
+        idx += chunk;
+        rem -= chunk;
+
         /* add up the bytes in this part */
         for ( ; chunk != 0 ; --chunk, ++p)
             hash += *p;
-
-        /* advance past the chunk */
-        idx += avail;
-        rem -= avail;
     }
 
     /* return the result */
@@ -1621,7 +1621,7 @@ int CVmObjByteArray::getp_read_int(VMG_ vm_obj_id_t self,
     static CVmNativeCodeDesc desc(2);
     unsigned int idx;
     unsigned int fmt;
-    long result;
+    long result = 0;
     size_t siz;
     unsigned char cbuf[4];
 
@@ -1844,6 +1844,10 @@ int CVmObjByteArray::write_to_file(osfildef *fp, unsigned long start_idx,
 {
     unsigned long rem;
     unsigned long idx;
+
+    /* make sure the starting index is valid */
+    if (start_idx < 1)
+        err_throw(VMERR_INDEX_OUT_OF_RANGE);
     
     /* 
      *   if the starting index is past the end of the array, there's nothing
@@ -1984,7 +1988,7 @@ int CVmObjByteArray::read_from_data_file(VMG_ vm_val_t *retval, osfildef *fp)
     /* read the number of bytes in the array */
     if (osfrb(fp, buf, 4))
         return 1;
-    len = osrp4(buf);
+    len = t3rp4u(buf);
 
     /* create a new ByteArray to hold the result */
     retval->set_obj(create(vmg_ FALSE, len));
