@@ -364,6 +364,27 @@ public:
         step_out_ = FALSE;
     }
 
+    /* 
+     *   Set single-step mode for a 'break' key (Ctrl-C, Ctrl+Break, etc.,
+     *   according to local conventions).  This asynchronously interrupts the
+     *   interpreter and breaks into the debugger while the program is
+     *   running, which is useful to break out of infinite loops or very
+     *   lengthy processing.  
+     */
+    void set_break_stop()
+    {
+        /* set single-step-in mode */
+        set_step_in();
+
+        /* 
+         *   Forget our last execution location, so that we'll stop again
+         *   even if we haven't moved from our last break location.  Since
+         *   we're explicitly breaking execution, we want to stop whether
+         *   we've gone anywhere or not.  
+         */
+        cur_stm_start_ = cur_stm_end_ = 0;
+    }
+
     /*
      *   Set single-step mode to STEP OVER.  In this mode, we will stop as
      *   soon as we reach a new statement at the same stack level as the
