@@ -84,6 +84,13 @@ enum vmobjfil_meta_fnset
     VMOBJFILE_OPEN_RES_RAW = 15
 };
 
+/* ------------------------------------------------------------------------ */
+/*
+ *   Special filename designators 
+ */
+
+/* library defaults file */
+#define SFID_LIB_DEFAULTS    0x0001
 
 /* ------------------------------------------------------------------------ */
 /*
@@ -250,7 +257,7 @@ int CVmObjFile::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                             vm_obj_id_t self, vm_obj_id_t *source_obj,
                             uint *argc)
 {
-    ushort func_idx;
+    uint func_idx;
 
     /* translate the property index to an index into our function table */
     func_idx = G_meta_table
@@ -345,7 +352,7 @@ void CVmObjFile::load_image_data(VMG_ const char *ptr, size_t siz)
     access = (unsigned char)*ptr++;
 
     /* get the flags */
-    flags = osrp4(ptr);
+    flags = t3rp4u(ptr);
 
     /* 
      *   add in the out-of-sync flag, since we've restored the state from a
@@ -465,6 +472,74 @@ void CVmObjFile::switch_read_write_mode(int writing)
         get_ext()->flags &= ~VMOBJFILE_LAST_OP_WRITE;
 }
 
+/* ------------------------------------------------------------------------ */
+/*
+ *   Retrieve the filename and access mode arguments. 
+ */
+void CVmObjFile::get_filename_and_access(VMG_ char *fname, size_t fname_siz,
+                                         int *access, int is_resource_file)
+{
+    int is_special_file = FALSE;
+    
+    /* 
+     *   check to see if we have an explicit filename string, or an integer
+     *   giving a special system file ID 
+     */
+    if (G_stk->get(0)->typ == VM_INT)
+    {
+        char path[OSFNMAX];
+
+        /* start with no path, in case we have trouble retrieving it */
+        path[0] = '\0';
+        
+        /* we have an integer, which is a special file designator */
+        switch (CVmBif::pop_int_val(vmg0_))
+        {
+        case SFID_LIB_DEFAULTS:
+            /* get the system application data path */
+            G_host_ifc->get_special_file_path(path, sizeof(path),
+                                              OS_GSP_T3_APP_DATA);
+
+            /* add the filename */
+            os_build_full_path(fname, fname_siz, path, "settings.txt");
+            break;
+
+        default:
+            /* invalid filename value */
+            err_throw(VMERR_BAD_VAL_BIF);
+        }
+
+        /* note that we have a special file, for file safety purposes */
+        is_special_file = TRUE;
+    }
+    else
+    {
+        /* we must have an explicit filename string - pop it */
+        CVmBif::pop_str_val_fname(vmg_ fname, fname_siz);
+    }
+
+    /* 
+     *   retrieve the access mode; if it's a resource file, the mode is
+     *   implicitly 'read' 
+     */
+    if (is_resource_file)
+        *access = VMOBJFILE_ACCESS_READ;
+    else
+        *access = CVmBif::pop_int_val(vmg0_);
+
+    /*
+     *   If this isn't a special file, then check the file safety mode to
+     *   ensure this operation is allowed.  Reading resources is always
+     *   allowed, regardless of the safety mode, since resources are
+     *   read-only and are inherently constrained in the paths they can
+     *   access.  Likewise, special files bypass the safety settings, because
+     *   the interpreter controls the names and locations of these files,
+     *   ensuring that they're inherently safe.  
+     */
+    if (!is_resource_file && !is_special_file)
+        check_safety_for_open(vmg_ fname, *access);
+}
+
 /* ------------------------------------------------------------------------ */
 /*
  *   Static property evaluator - open a text file 
@@ -501,16 +576,8 @@ int CVmObjFile::s_getp_open_text(VMG_ vm_val_t *retval, uint *in_argc,
     res_end = 0;
 
     /* retrieve the filename */
-    CVmBif::pop_str_val_fname(vmg_ fname, sizeof(fname));
-
-    /* 
-     *   retrieve the access mode; if it's a resource file, the mode is
-     *   implicitly 'read' 
-     */
-    if (is_resource_file)
-        access = VMOBJFILE_ACCESS_READ;
-    else
-        access = CVmBif::pop_int_val(vmg0_);
+    get_filename_and_access(vmg_ fname, sizeof(fname),
+                            &access, is_resource_file);
 
     /* presume we won't need a read buffer */
     create_readbuf = FALSE;
@@ -555,15 +622,6 @@ int CVmObjFile::s_getp_open_text(VMG_ vm_val_t *retval, uint *in_argc,
     /* push the character map object onto the stack for gc protection */
     G_stk->push()->set_obj(cset_obj);
 
-    /* 
-     *   Check the file safety mode to ensure this operation is allowed.
-     *   Reading resources is always allowed, regardless of the safety mode,
-     *   since resources are read-only and are inherently constrained in the
-     *   paths they can access.  
-     */
-    if (!is_resource_file)
-        check_safety_for_open(vmg_ fname, access);
-
     /* open the file for reading or writing, as appropriate */
     switch(access)
     {
@@ -645,6 +703,7 @@ int CVmObjFile::s_getp_open_text(VMG_ vm_val_t *retval, uint *in_argc,
         break;
 
     default:
+        fp = 0;
         err_throw(VMERR_BAD_VAL_BIF);
     }
 
@@ -711,25 +770,9 @@ int CVmObjFile::open_binary(VMG_ vm_val_t *retval, uint *argc, int mode,
     res_start = 0;
     res_end = 0;
 
-    /* retrieve the filename */
-    CVmBif::pop_str_val_fname(vmg_ fname, sizeof(fname));
-
-    /* 
-     *   retrieve the access mode - if it's a resource file, the mode is
-     *   implicitly 'read' 
-     */
-    if (is_resource_file)
-        access = VMOBJFILE_ACCESS_READ;
-    else
-        access = CVmBif::pop_int_val(vmg0_);
-
-    /* 
-     *   check the file safety mode to ensure this operation is allowed (but
-     *   we don't have to do this for resource files, since resource files
-     *   are always readable regardless of the safety mode) 
-     */
-    if (!is_resource_file)
-        check_safety_for_open(vmg_ fname, access);
+    /* retrieve the filename and access mode */
+    get_filename_and_access(vmg_ fname, sizeof(fname),
+                            &access, is_resource_file);
 
     /* open the file in binary mode, with the desired access type */
     switch(access)
@@ -800,6 +843,7 @@ int CVmObjFile::open_binary(VMG_ vm_val_t *retval, uint *argc, int mode,
         break;
 
     default:
+        fp = 0;
         err_throw(VMERR_BAD_VAL_BIF);
     }
 
@@ -843,11 +887,11 @@ void CVmObjFile::check_safety_for_open(VMG_ const char *fname, int access)
         /*
          *   we want only read access - we can't read at all if the safety
          *   level isn't READ_CUR or below, and we must be at level
-         *   WRITE_NONE or lower to read from a file not in the current
-         *   directory 
+         *   READ_ANY_WRITE_CUR or lower to read from a file not in the
+         *   current directory 
          */
         if (safety > VM_IO_SAFETY_READ_CUR
-            || (!in_same_dir && safety > VM_IO_SAFETY_WRITE_NONE))
+            || (!in_same_dir && safety > VM_IO_SAFETY_READ_ANY_WRITE_CUR))
         {
             /* this operation is not allowed - throw an error */
             G_interpreter->throw_new_class(vmg_ G_predef->file_safety_exc,
@@ -859,11 +903,11 @@ void CVmObjFile::check_safety_for_open(VMG_ const char *fname, int access)
     case VMOBJFILE_ACCESS_RW_KEEP:
     case VMOBJFILE_ACCESS_RW_TRUNC:
         /* 
-         *   writing - we must be safety level below WRITE_NONE to write at
-         *   all, and we must be at level MINIMUM to write a file that's not
-         *   in the current directory 
+         *   writing - we must be safety level of at least READWRITE_CUR to
+         *   write at all, and we must be at level MINIMUM to write a file
+         *   that's not in the current directory 
          */
-        if (safety >= VM_IO_SAFETY_WRITE_NONE
+        if (safety > VM_IO_SAFETY_READWRITE_CUR
             || (!in_same_dir && safety > VM_IO_SAFETY_MINIMUM))
         {
             /* this operation is not allowed - throw an error */
@@ -1290,7 +1334,7 @@ void CVmObjFile::read_data_mode(VMG_ vm_val_t *retval)
             goto io_error;
 
         /* set the 'enum' value */
-        retval->set_enum(osrp4(buf));
+        retval->set_enum(t3rp4u(buf));
         break;
 
     case VMOBJFILE_TAG_STRING:
