@@ -110,15 +110,17 @@ private:
 /*
  *   instantiation 
  */
-CVmImageLoader::CVmImageLoader(CVmImageFile *fp, const char *fname)
+CVmImageLoader::CVmImageLoader(CVmImageFile *fp, const char *fname,
+                               long base_ofs)
 {
     size_t i;
     
     /* remember the underlying image file interface */
     fp_ = fp;
 
-    /* remember the image filename */
+    /* remember the image filename and the base seek offset */
     fname_ = lib_copy_str(fname);
+    base_seek_ofs_ = base_ofs;
 
     /* we don't know the file version yet */
     ver_ = 0;
@@ -234,7 +236,7 @@ void CVmImageLoader::load(VMG0_)
         fp_->copy_data(buf, 10);
 
         /* get the size */
-        siz = osrp4(buf + 4);
+        siz = t3rp4u(buf + 4);
 
         /* get the flags */
         flags = osrp2(buf + 8);
@@ -257,7 +259,7 @@ void CVmImageLoader::load(VMG0_)
         }
         else if (block_type_is(buf, "CPDF"))
         {
-            /* load the constant pool definiton block */
+            /* load the constant pool definition block */
             load_const_pool_def(siz);
         }
         else if (block_type_is(buf, "MRES"))
@@ -356,6 +358,9 @@ void CVmImageLoader::load(VMG0_)
     /* complete the dynamic linking */
     do_dynamic_link(vmg0_);
 
+    /* fix up the global symbol table metaclasses, if applicable */
+    fix_gsym_meta(vmg0_);
+
     /* 
      *   create an IntrinsicClass instance for each metaclass that the
      *   image file is using and for which the compiler didn't supply its
@@ -474,7 +479,7 @@ void CVmImageLoader::load_ext_resfiles(VMG0_)
 
                 /* set up the loader */
                 imagefp = new CVmImageFileExt(fp);
-                loader = new CVmImageLoader(imagefp, resfile);
+                loader = new CVmImageLoader(imagefp, resfile, 0);
 
                 /* load the resource-only file */
                 loader->load_resource_file(&res_ifc);
@@ -543,7 +548,7 @@ void CVmImageLoader::load_resources_from_fp(osfildef *fp,
     {
         /* set up the loader */
         imagefp = new CVmImageFileExt(file);
-        loader = new CVmImageLoader(imagefp, fname);
+        loader = new CVmImageLoader(imagefp, fname, 0);
 
         /* load the resource-only file */
         loader->load_resource_file(res_ifc);
@@ -590,7 +595,7 @@ void CVmImageLoader::load_resource_file(CVmImageLoaderMres *res_ifc)
         fp_->copy_data(buf, 10);
 
         /* get the size */
-        siz = osrp4(buf + 4);
+        siz = t3rp4u(buf + 4);
 
         /* get the flags */
         flags = osrp2(buf + 8);
@@ -1000,7 +1005,7 @@ void CVmImageLoader::load_entrypt(VMG_ ulong siz)
     read_data(buf, 16, &siz);
 
     /* set the entrypoint */
-    entrypt_ = osrp4(buf);
+    entrypt_ = t3rp4u(buf);
 
     /* set the method header size in the interpreter */
     G_interpreter->set_funchdr_size(osrp2(buf+4));
@@ -1040,8 +1045,8 @@ void CVmImageLoader::load_const_pool_def(ulong siz)
 
     /* decode the pool identifier, page count, and page size */
     pool_id = osrp2(buf);
-    page_count = osrp4(buf + 2);
-    page_size = osrp4(buf + 6);
+    page_count = t3rp4u(buf + 2);
+    page_size = t3rp4u(buf + 6);
 
     /* ensure that the pool ID is valid */
     if (pool_id < 1 || pool_id > sizeof(pools_)/sizeof(pools_[0]))
@@ -1071,7 +1076,7 @@ void CVmImageLoader::load_const_pool_page(ulong siz)
 
     /* decode the pool ID and page index */
     pool_id = osrp2(buf);
-    idx = osrp4(buf + 2);
+    idx = t3rp4u(buf + 2);
     xor_mask = buf[6];
 
     /* ensure that the pool ID is valid */
@@ -1135,7 +1140,7 @@ void CVmImageLoader::load_static_objs(VMG_ ulong siz)
         read_data(buf, header_size, &siz);
 
         /* get the object ID */
-        obj_id = (vm_obj_id_t)osrp4(buf);
+        obj_id = (vm_obj_id_t)t3rp4u(buf);
 
         /* get the size */
         if (large_objs)
@@ -1144,11 +1149,11 @@ void CVmImageLoader::load_static_objs(VMG_ ulong siz)
              *   make sure the object size doesn't overflow the local
              *   hardware limits 
              */
-            if ((unsigned long)osrp4(buf + 4) > (unsigned long)OSMALMAX)
+            if (t3rp4u(buf + 4) > (unsigned long)OSMALMAX)
                 err_throw(VMERR_OBJ_SIZE_OVERFLOW);
             
             /* read the object size */
-            obj_size = (size_t)osrp4(buf + 4);
+            obj_size = (size_t)t3rp4u(buf + 4);
         }
         else
         {
@@ -1181,10 +1186,16 @@ void CVmImageLoader::load_mres(ulong siz, CVmImageLoaderMres *res_ifc)
     uint i;
 
     /* 
-     *   note the current seek position - each entry's seek position is
-     *   stored in the table of contents as an offset from this location 
+     *   Note the current physical seek position - each entry's seek position
+     *   is stored in the table of contents as an offset from this location.
+     *   
+     *   Compute the physical base offset: this is the logical base offset in
+     *   our image stream, plus the offset of the image stream within the
+     *   image file.  We need the physical base offset because the underlying
+     *   interpreter's resource loader works with the physical file, not the
+     *   logical image stream.  
      */
-    base_ofs = fp_->get_seek();
+    base_ofs = fp_->get_seek() + base_seek_ofs_;
     
     /* read the entry count and size of the table of contents */
     read_data(buf, 2, &siz);
@@ -1202,8 +1213,8 @@ void CVmImageLoader::load_mres(ulong siz, CVmImageLoaderMres *res_ifc)
 
         /* read the fixed part of the table-of-contents entry */
         read_data(buf, 9, &siz);
-        entry_ofs = osrp4(buf);
-        entry_size = osrp4(buf + 4);
+        entry_ofs = t3rp4u(buf);
+        entry_size = t3rp4u(buf + 4);
         entry_name_len = (uchar)*(buf + 8);
 
         /* read the name */
@@ -1261,7 +1272,7 @@ void CVmImageLoader::load_meta_dep(VMG_ ulong siz)
     /* read the entries */
     for (i = 0 ; i < entry_cnt ; ++i)
     {
-        uint j;
+        ushort j;
         uint len;
         uint prop_cnt;
         uint prop_len;
@@ -1492,7 +1503,7 @@ void CVmImageLoader::load_runtime_symtab_from_gsym(VMG_ ulong siz)
 
     /* read the symbol count */
     read_data(buf, 4, &siz);
-    cnt = osrp4(buf);
+    cnt = t3rp4u(buf);
 
     /* read the symbols and populate the symbol table */
     for ( ; cnt != 0 ; --cnt)
@@ -1551,12 +1562,12 @@ void CVmImageLoader::load_runtime_symtab_from_gsym(VMG_ ulong siz)
         {
         case TC_SYM_FUNC:
             /* set up the function pointer value */
-            val.set_fnptr(osrp4(dat));
+            val.set_fnptr(t3rp4u(dat));
             break;
 
         case TC_SYM_OBJ:
             /* set up the object value */
-            val.set_obj(osrp4(dat));
+            val.set_obj(t3rp4u(dat));
             break;
 
         case TC_SYM_PROP:
@@ -1566,12 +1577,12 @@ void CVmImageLoader::load_runtime_symtab_from_gsym(VMG_ ulong siz)
 
         case TC_SYM_ENUM:
             /* set up the enum value */
-            val.set_enum(osrp4(dat));
+            val.set_enum(t3rp4u(dat));
             break;
 
         case TC_SYM_METACLASS:
             /* set up the metaclass object value */
-            val.set_obj(osrp4(dat + 2));
+            val.set_obj(t3rp4u(dat + 2));
             break;
 
         default:
@@ -1647,7 +1658,7 @@ void CVmImageLoader::load_srcfiles(VMG_ ulong siz)
          *   calculate the file location of the start of the next block by
          *   adding the starting position and size of this block 
          */
-        next_pos = start_pos + osrp4(buf);
+        next_pos = start_pos + t3rp4u(buf);
 
         /* allocate the entry */
         entry = G_srcf_table->add_entry(orig_index, len);
@@ -1660,7 +1671,7 @@ void CVmImageLoader::load_srcfiles(VMG_ ulong siz)
 
         /* read the number of line records */
         read_data(buf, 4, &siz);
-        line_cnt = osrp4(buf);
+        line_cnt = t3rp4u(buf);
 
         /* 
          *   if the line records are too big for our buffer, or smaller than
@@ -1682,8 +1693,8 @@ void CVmImageLoader::load_srcfiles(VMG_ ulong siz)
                 read_data(buf, line_size, &siz);
 
                 /* get the source line number and byte code offset */
-                linenum = osrp4(buf);
-                code_addr = osrp4(buf + 4);
+                linenum = t3rp4u(buf);
+                code_addr = t3rp4u(buf + 4);
 
                 /* add the line to the source file table entry */
                 entry->add_line_record(linenum, code_addr);
@@ -1889,9 +1900,9 @@ void CVmImageLoader::load_sini(VMG_ ulong siz)
      *   initializer count 
      */
     fp_->copy_data(hdr, 12);
-    hdr_siz = osrp4(hdr);
-    static_cs_ofs_ = osrp4(hdr + 4);
-    init_cnt = osrp4(hdr + 8);
+    hdr_siz = t3rp4u(hdr);
+    static_cs_ofs_ = t3rp4u(hdr + 4);
+    init_cnt = t3rp4u(hdr + 8);
 
     /* skip any extra header information */
     if (hdr_siz > 12)
@@ -2051,7 +2062,7 @@ int CVmImageLoader::restore_synth_exports(VMG_ CVmFile *fp,
     G_image_loader->discard_synth_exports();
 
     /* read the number of synthesized exports */
-    cnt = fp->read_int4();
+    cnt = fp->read_uint4();
 
     /* read the exports */
     for ( ; cnt != 0 ; --cnt)
