@@ -424,7 +424,7 @@ protected:
     vm_lookup_val *alloc_new_entry(VMG0_);
 
     /* calculate a value's hash code */
-    uint calc_hash(VMG_ const vm_val_t *key);
+    uint calc_key_hash(VMG_ const vm_val_t *key);
 
     /* get my extension data */
     vm_lookup_ext *get_ext() const { return (vm_lookup_ext *)ext_; }
