@@ -230,14 +230,10 @@ public:
     vm_val_t *get_r0() { return &r0_; }
 
     /* set the default "say" function */
-    void set_say_func(pool_ofs_t code_pool_ofs)
-    {
-        /* set the offset */
-        say_func_ = code_pool_ofs;
+    void set_say_func(VMG_ const vm_val_t *val);
 
-        /* note that it's been initialized */
-        say_func_inited_ = TRUE;
-    }
+    /* get the current default "say" function */
+    void get_say_func(vm_val_t *val) const;
 
     /* set the default "say" method */
     void set_say_method(vm_prop_id_t prop)
@@ -246,6 +242,9 @@ public:
         say_method_ = prop;
     }
 
+    /* get the current "say" method */
+    vm_prop_id_t get_say_method() const { return say_method_; }
+
     /* pop an integer value; throws an error if the value is not an integer */
     void pop_int(VMG_ vm_val_t *val)
     {
@@ -548,7 +547,10 @@ public:
     
     /*
      *   Get the current "self" object.  The "self" object is always the
-     *   implicit first parameter to any method.  
+     *   implicit first parameter to any method.  Note that this version of
+     *   the method *doesn't* check for nil - it assumes that the caller
+     *   knows for sure that there's a valid "self", so dispenses with any
+     *   checks to save time.  
      */
     vm_obj_id_t get_self(VMG0_) const
     {
@@ -556,6 +558,21 @@ public:
         return G_stk->get_from_frame(frame_ptr_, VMRUN_FPOFS_SELF)->val.obj;
     }
 
+    /*
+     *   Get the current "self" object, checking for nil.  If "self" is nil,
+     *   we'll return VM_INVALID_OBJ.  This version (not get_self()) should
+     *   be used whenever it's not certain from context that there's a valid
+     *   "self".  
+     */
+    vm_obj_id_t get_self_check(VMG0_) const
+    {
+        /* get the 'self' slot from the stack frame */
+        vm_val_t *valp = G_stk->get_from_frame(frame_ptr_, VMRUN_FPOFS_SELF);
+
+        /* if it's nil, return VM_INVALID_OBJ; otherwise, return the obj ID */
+        return (valp->typ == VM_NIL ? VM_INVALID_OBJ : valp->val.obj);
+    }
+
     /* set the current 'self' object */
     void set_self(VMG_ const vm_val_t *val)
     {
@@ -1068,7 +1085,7 @@ protected:
     /* get an unsigned 32-bit byte-code operand */
     uint32 get_op_uint32(const uchar **p)
     {
-        uint32 ret = (uint32)osrp4(*p);
+        uint32 ret = (uint32)t3rp4u(*p);
         *p += 4;
         return ret;
     }
@@ -1116,12 +1133,12 @@ protected:
     size_t funchdr_size_;
 
     /*
-     *   The function to invoke for the SAY and SAYVAL opcodes.  This is
-     *   an offset into the code pool.  If say_func_inited_ is not set,
-     *   the SAY opcode will throw an error.  
+     *   A pointer to a global variable in the object table (CVmObjTable)
+     *   containing the function to invoke for the SAY and SAYVAL opcodes.
+     *   This can be a function pointer, a function object, or nil.  If this
+     *   is nil, the SAY opcode will throw an error.  
      */
-    pool_ofs_t say_func_;
-    int say_func_inited_;
+    struct vm_globalvar_t *say_func_;
 
     /*
      *   The method to invoke for the SAY and SAYVAL opcodes when a valid
