@@ -340,7 +340,7 @@ vm_tadsobj_hdr *vm_tadsobj_hdr::alloc(VMG_ CVmObjTads *self,
                                       unsigned short sc_cnt,
                                       unsigned short prop_cnt)
 {
-    size_t hash_siz;
+    ushort hash_siz;
     size_t siz;
     size_t i;
     vm_tadsobj_hdr *hdr;
@@ -468,7 +468,7 @@ vm_tadsobj_hdr *vm_tadsobj_hdr::expand_to(VMG_ CVmObjTads *self,
     vm_tadsobj_prop *entryp;
 
     /* allocate a new object at the expanded property table size */
-    new_hdr = alloc(vmg_ self, new_sc_cnt, new_prop_cnt);
+    new_hdr = alloc(vmg_ self, (ushort)new_sc_cnt, (ushort)new_prop_cnt);
 
     /* copy the superclasses from the original object */
     memcpy(new_hdr->sc, hdr->sc,
@@ -491,9 +491,10 @@ vm_tadsobj_hdr *vm_tadsobj_hdr::expand_to(VMG_ CVmObjTads *self,
     new_hdr->inh_path = hdr->inh_path;
 
     /* 
-     *   Run through all of the existing properties and duplicate them in
-     *   the new object, to build the new object's hash table.  Note that
-     *   the free index is also the count of properties in use.  
+     *   Run through all of the existing properties and duplicate them in the
+     *   new object, to build the new object's hash table.  Note that the
+     *   free index is inherently equivalent to the count of properties in
+     *   use.  
      */
     for (i = hdr->prop_entry_free, entryp = hdr->prop_entry_arr ; i != 0 ;
          --i, ++entryp)
@@ -786,7 +787,7 @@ vm_obj_id_t CVmObjTads::create_from_stack_multi(
 {
     vm_obj_id_t id;
     CVmObjTads *obj;
-    uint i;
+    ushort i;
 
     /* allocate an object ID */
     id = vm_new_id(vmg_ FALSE, TRUE, FALSE);
@@ -794,7 +795,7 @@ vm_obj_id_t CVmObjTads::create_from_stack_multi(
         G_obj_table->set_obj_transient(id);
 
     /* create the new object */
-    obj = new (vmg_ id) CVmObjTads(vmg_ argc, VMTOBJ_PROP_INIT);
+    obj = new (vmg_ id) CVmObjTads(vmg_ (ushort)argc, VMTOBJ_PROP_INIT);
 
     /* push the new object, for garbage collector protection */
     G_interpreter->push_obj(vmg_ id);
@@ -894,6 +895,9 @@ vm_obj_id_t CVmObjTads::create_from_stack_multi(
                                 &new_obj_val, lst_cnt - 1);
     }
 
+    /* discard the arguments plus our own gc protection */
+    G_stk->discard(argc + 1);
+
     /* return the new object */
     return id;
 }
@@ -1047,18 +1051,22 @@ void CVmObjTads::set_prop(VMG_ CVmUndo *undo, vm_obj_id_t self,
     /* look for an existing property entry */
     entry = hdr->find_prop_entry(prop);
 
-    /* if we didn't find an existing entry, allocate a new one */
+    /* check for an existing entry for the property */
     if (entry != 0)
     {
-        /* note the old value */
+        /* found an existing entry - note the old value */
         oldval = entry->val;
 
-        /* store the new value */
+        /* store the new value in the existing entry */
         entry->val = *val;
     }
     else
     {
-        /* if we don't have any free properties left, reallocate */
+        /* 
+         *   We didn't find an existing entry for the property, so we have to
+         *   add a new one.  If we don't have any free property slots left,
+         *   expand the object to create some more property slots.  
+         */
         if (!hdr->has_free_entries(1))
         {
             /* expand the extension to make room for more properties */
@@ -1540,8 +1548,8 @@ int CVmObjTads::get_prop_intrinsic(VMG_ vm_prop_id_t prop, vm_val_t *val,
                                    vm_obj_id_t self, vm_obj_id_t *source_obj,
                                    uint *argc)
 {
-    unsigned short func_idx;
-
+    uint func_idx;
+    
     /* translate the property into a function vector index */
     func_idx = G_meta_table
                ->prop_to_vector_idx(metaclass_reg_->get_reg_idx(), prop);
@@ -1755,6 +1763,24 @@ void CVmObjTads::apply_undo(VMG_ CVmUndoRecord *rec)
     vm_tadsobj_prop *entry;
     vm_tadsobj_hdr *hdr = get_hdr();
 
+    /* 
+     *   if the property is 'invalid', this is an undo record for a
+     *   superclass list change rather than a property change 
+     */
+    if (rec->id.prop == VM_INVALID_PROP)
+    {
+        const char *lstp;
+
+        /* get the old list */
+        lstp = rec->oldval.get_as_list(vmg0_);
+
+        /* set the new superclass list */
+        change_superclass_list(vmg_ lstp, (ushort)vmb_get_len(lstp));
+
+        /* we're done with this undo record */
+        return;
+    }
+
     /* find the property entry for the property being undone */
     entry = hdr->find_prop_entry(rec->id.prop);
     if (entry == 0)
@@ -1951,10 +1977,10 @@ void CVmObjTads::restore_from_file(VMG_ vm_obj_id_t self,
     vm_tadsobj_hdr *hdr;
     
     /* read number of modified properties */
-    mod_count = fp->read_uint2();
+    mod_count = (ushort)fp->read_uint2();
 
     /* read the number of superclasses */
-    sc_cnt = fp->read_uint2();
+    sc_cnt = (ushort)fp->read_uint2();
 
     /* 
      *   If we don't have an extension yet, allocate one.  The only way we
@@ -1996,7 +2022,7 @@ void CVmObjTads::restore_from_file(VMG_ vm_obj_id_t self,
         vm_obj_id_t sc;
 
         /* read the next superclass */
-        sc = (vm_obj_id_t)fp->read_int4();
+        sc = (vm_obj_id_t)fp->read_uint4();
 
         /* fix it up to the memory numbering system */
         sc = fixups->get_new_id(vmg_ sc);
@@ -2141,7 +2167,7 @@ void CVmObjTads::load_image_props_and_scs(VMG_ const char *ptr, size_t siz)
     for (i = 0, p = ptr + 6 ; i < sc_cnt ; ++i, p += 4)
     {
         /* store the object ID */
-        hdr->sc[i].id = (vm_obj_id_t)osrp4(p);
+        hdr->sc[i].id = (vm_obj_id_t)t3rp4u(p);
 
         /* 
          *   we can't store the superclass pointer yet, as the superclass
@@ -2231,7 +2257,7 @@ int CVmObjTads::getp_create_clone(VMG_ vm_obj_id_t self,
     vm_obj_id_t new_obj;
     CVmObjTads *tobj;
     vm_tadsobj_prop *entry;
-    size_t i;
+    ushort i;
     vm_tadsobj_hdr *hdr = get_hdr();
 
     /* check arguments */
@@ -2321,10 +2347,10 @@ int CVmObjTads::getp_set_sc_list(VMG_ vm_obj_id_t self,
 {
     static CVmNativeCodeDesc desc(1);
     const char *lstp;
-    size_t cnt;
+    ushort cnt;
     size_t i;
     vm_val_t ele;
-    size_t sc_cnt;
+    ushort sc_cnt;
     vm_tadsobj_hdr *hdr = get_hdr();
 
     /* check arguments */
@@ -2337,7 +2363,7 @@ int CVmObjTads::getp_set_sc_list(VMG_ vm_obj_id_t self,
         err_throw(VMERR_BAD_TYPE_BIF);
 
     /* get the number of superclasses for the new object */
-    cnt = vmb_get_len(lstp);
+    cnt = (ushort)vmb_get_len(lstp);
 
     /* we need at least one argument - the minimal root is TadsObject */
     if (cnt < 1)
@@ -2386,26 +2412,79 @@ int CVmObjTads::getp_set_sc_list(VMG_ vm_obj_id_t self,
         sc_cnt = cnt;
     }
 
+    /* if there's a system undo object, add undo for the change */
+    if (G_undo != 0)
+    {
+        vm_val_t oldv;
+        CVmObjList *oldp;
+
+        /* allocate a list for the results */
+        oldv.set_obj(CVmObjList::create(vmg_ FALSE, hdr->sc_cnt));
+        oldp = (CVmObjList *)vm_objp(vmg_ oldv.val.obj);
+
+        /* build the superclass list */
+        for (i = 0 ; i < hdr->sc_cnt ; ++i)
+        {
+            /* add this superclass to the list */
+            ele.set_obj(hdr->sc[i].id);
+            oldp->cons_set_element(i, &ele);
+        }
+
+        /* 
+         *   Add an undo record with the original superclass list as the old
+         *   value.  Use the 'invalid' property as the proprety key - all of
+         *   our other undo records are associated with actual properties, so
+         *   this is how we know this is an undo record for the superclass
+         *   list.  
+         */
+        G_undo->add_new_record_prop_key(vmg_ self, VM_INVALID_PROP, &oldv);
+    }
+
+    /* update the superclass list with the given list */
+    change_superclass_list(vmg_ lstp, sc_cnt);
+
+    /* discard arguments */
+    G_stk->discard();
+
+    /* no return value */
+    retval->set_nil();
+
+    /* handled */
+    return TRUE;
+}
+
+/*
+ *   Change the superclass list to the given list.  'lstp' is the new
+ *   superclass list, in constant list format (i.e., a packed array of
+ *   dataholder values).  
+ */
+void CVmObjTads::change_superclass_list(VMG_ const char *lstp, ushort cnt)
+{
+    vm_tadsobj_hdr *hdr = get_hdr();
+    size_t i;
+
     /* 
      *   if we're increasing the number of superclasses, expand our object
      *   header to make room 
      */
-    if (sc_cnt > hdr->sc_cnt)
+    if (cnt > hdr->sc_cnt)
     {
         /* expand the header to accomodate the new superclass list */
         ext_ = (char *)vm_tadsobj_hdr::expand_to(
-            vmg_ this, hdr, sc_cnt, hdr->prop_entry_cnt);
+            vmg_ this, hdr, cnt, hdr->prop_entry_cnt);
 
         /* get the new header */
         hdr = get_hdr();
     }
 
     /* set the new superclass count */
-    hdr->sc_cnt = sc_cnt;
+    hdr->sc_cnt = cnt;
 
     /* set the new superclasses */
-    for (i = 0 ; i < sc_cnt ; ++i)
+    for (i = 0 ; i < cnt ; ++i)
     {
+        vm_val_t ele;
+
         /* get this element from the list */
         CVmObjList::index_list(vmg_ &ele, lstp, i + 1);
 
@@ -2416,12 +2495,6 @@ int CVmObjTads::getp_set_sc_list(VMG_ vm_obj_id_t self,
 
     /* invalidate the cached inheritance path */
     hdr->inval_inh_path();
-
-    /* discard arguments */
-    G_stk->discard();
-
-    /* handled */
-    return TRUE;
 }
 
 /* ------------------------------------------------------------------------ */
