@@ -141,7 +141,7 @@ vm_obj_id_t CVmObjVector::create_from_stack(VMG_ const uchar **pc_ptr,
     vm_val_t *arg1;
     vm_obj_id_t id;
     CVmObjVector *vec;
-    size_t cnt;
+    size_t cnt = 0;
     size_t copy_cnt;
     const char *src_lst;
     CVmObjVector *src_vec;
@@ -352,6 +352,21 @@ vm_obj_id_t CVmObjVector::create_copy(VMG_ const vm_val_t *self_val)
     return new_id;
 }
 
+/* ------------------------------------------------------------------------ */
+/*
+ *   Set the allocation size - this is the number of elements we've allocated
+ *   space for, which might exceed the number of elements we're actually
+ *   storing currently.  
+ */
+void CVmObjVector::set_allocated_count(size_t cnt)
+{
+    /* if the new size exceeds the maximum allowed size, throw an error */
+    if (cnt > 65535)
+        err_throw(VMERR_OUT_OF_RANGE);
+
+    /* set the new allocated size */
+    vmb_put_len(cons_get_vector_ext_ptr(), cnt);
+}
 
 /* ------------------------------------------------------------------------ */
 /*
@@ -370,6 +385,9 @@ void CVmObjVector::new_iterator(VMG_ vm_val_t *retval,
      */
     copy_val.set_obj(create_copy(vmg_ self_val));
 
+    /* push the copy for gc protection */
+    G_stk->push(&copy_val);
+
     /* 
      *   Set up a new indexed iterator object.  The first valid index for a
      *   vector is always 1, and the last valid index is the same as the
@@ -377,6 +395,9 @@ void CVmObjVector::new_iterator(VMG_ vm_val_t *retval,
      */
     retval->set_obj(CVmObjIterIdx::create_for_coll(vmg_ &copy_val,
         1, get_element_count()));
+
+    /* discard gc protection */
+    G_stk->discard();
 }
 
 /*
@@ -414,7 +435,7 @@ int CVmObjVector::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                            vm_obj_id_t self, vm_obj_id_t *source_obj,
                            uint *argc)
 {
-    ushort func_idx;
+    uint func_idx;
 
     /* translate the property index to an index into our function table */
     func_idx = G_meta_table
@@ -860,6 +881,9 @@ int CVmObjVector::equals(VMG_ vm_obj_id_t self, const vm_val_t *val,
 
         /* get the number of elements in the list */
         cnt2 = vmb_get_len(p);
+
+        /* the second element isn't a vector */
+        vec2 = 0;
     }
     else
     {
@@ -1209,14 +1233,14 @@ int CVmObjVector::getp_copy_from(VMG_ vm_obj_id_t self, vm_val_t *retval,
                                  uint *argc)
 {
     vm_val_t src_val;
-    size_t src_cnt;
+    size_t src_cnt = 0;
     size_t dst_cnt;
     size_t src_start;
     size_t dst_start;
     size_t copy_cnt;
     const char *src_lst;
-    int src_is_list;
-    CVmObjVector *src_vec;
+    int src_is_list = FALSE;
+    CVmObjVector *src_vec = 0;
     size_t i;
     static CVmNativeCodeDesc desc(4);
 
@@ -1255,9 +1279,6 @@ int CVmObjVector::getp_copy_from(VMG_ vm_obj_id_t self, vm_val_t *retval,
         /* note that it's a list */
         src_is_list = TRUE;
 
-        /* there's no source vector */
-        src_vec = 0;
-
         /* note the source size */
         src_cnt = vmb_get_len(src_lst);
     }
@@ -1268,9 +1289,6 @@ int CVmObjVector::getp_copy_from(VMG_ vm_obj_id_t self, vm_val_t *retval,
         /* it's a vector */
         src_vec = (CVmObjVector *)vm_objp(vmg_ src_val.val.obj);
 
-        /* it's not a list */
-        src_is_list = FALSE;
-
         /* note the soure size */
         src_cnt = src_vec->get_element_count();
     }
@@ -1284,8 +1302,8 @@ int CVmObjVector::getp_copy_from(VMG_ vm_obj_id_t self, vm_val_t *retval,
         copy_cnt = src_cnt - src_start;
 
     /* expand the vector if necessary to make room for added elements */
-    if (dst_start + copy_cnt - 1 > dst_cnt)
-        set_element_count_undo(vmg_ self, dst_start + copy_cnt - 1);
+    if (dst_start + copy_cnt > dst_cnt)
+        set_element_count_undo(vmg_ self, dst_start + copy_cnt);
 
     /* set the list elements */
     for (i = 0 ; i < copy_cnt ; ++i)
@@ -1787,7 +1805,7 @@ int CVmObjVector::for_each_gen(VMG_ vm_obj_id_t self,
 
         /* push the index argument if desired */
         if (pass_key_to_cb)
-            G_stk->push()->set_int(idx);
+            G_stk->push()->set_int(idx + 1);
 
         /* invoke the callback */
         G_interpreter->call_func_ptr(vmg_ func_val, pass_key_to_cb ? 2 : 1,
@@ -2328,6 +2346,10 @@ int CVmObjVector::getp_set_length(VMG_ vm_obj_id_t self, vm_val_t *retval,
     /* get the new length */
     new_len = (size_t)CVmBif::pop_int_val(vmg0_);
 
+    /* can't go less than zero */
+    if (new_len < 0)
+        err_throw(VMERR_BAD_VAL_BIF);
+
     /* set the vector to its new size */
     set_element_count_undo(vmg_ self, new_len);
 
